# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from importlib import import_module


def import_object(class_name: str):
    """Imports an object from a module based on a string

    Args:
        class_name (str): The full path to the object e.g. "package.module.Class"

    Returns:
        object: The imported object
    """
    module_path, class_name = class_name.rsplit(".", 1)
    module = import_module(module_path)
    return getattr(module, class_name)



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import SimpleNamespace

import cogment as _cog

import cogment_lab.generated.data_pb2 as data_pb
import cogment_lab.generated.ndarray_pb2 as ndarray_pb
import cogment_lab.generated.spaces_pb2 as spaces_pb


_player_class = _cog.actor.ActorClass(
    name="player",
    config_type=data_pb.AgentConfig,
    action_space=data_pb.PlayerAction,
    observation_space=data_pb.Observation,
)


actor_classes = _cog.actor.ActorClassList(_player_class)

trial = SimpleNamespace(config_type=data_pb.TrialConfig)

environment = SimpleNamespace(config_type=data_pb.EnvironmentConfig)



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import gymnasium as gym

from cogment_lab.generated.data_pb2 import AgentSpecs as PbAgentSpecs  # type: ignore

from ..constants import DEFAULT_RENDERED_WIDTH
from .action_space import ActionSpace
from .ndarray_serialization import SerializationFormat
from .observation_space import ObservationSpace
from .spaces_serialization import deserialize_space, serialize_gym_space


class AgentSpecs:
    """
    Representation of the specification of an agent within Cogment Lab.
    """

    def __init__(self, agent_specs_pb: PbAgentSpecs):
        """
        AgentSpecs constructor.
        Shouldn't be called directly, prefer the factory function such as AgentSpecs.deserialize or AgentSpecs.create_homogeneous.
        """
        self._pb = agent_specs_pb

    def get_observation_space(self, render_width: int = DEFAULT_RENDERED_WIDTH) -> ObservationSpace:
        """
        Build an instance of the observation space for this agent

        Parameters:
            render_width: optional
                maximum width for the serialized rendered frame in observation

        NOTE: In the future we'll want to support different observation space per agent role
        """
        return ObservationSpace(deserialize_space(self._pb.observation_space), render_width)

    def get_action_space(self, seed: int | None = None) -> ActionSpace:
        """
        Build an instance of the action space for this agent

        Parameters:
            seed: optional
                the seed used when generating random actions

        NOTE: In the future we'll want to support different action space per agent roles
        """
        return ActionSpace(deserialize_space(self._pb.action_space), seed)

    @classmethod
    def create_homogeneous(
        cls,
        observation_space: gym.Space,
        action_space: gym.Space,
        serialization_format: SerializationFormat = SerializationFormat.STRUCTURED,
    ):
        """
        Factory function building an AgentSpecs.
        """
        return cls.deserialize(
            PbAgentSpecs(
                observation_space=serialize_gym_space(observation_space, serialization_format),
                action_space=serialize_gym_space(action_space, serialization_format),
            )
        )

    def serialize(self):
        """
        Serialize to a AgentSpecs protobuf message
        """
        return self._pb

    @classmethod
    def deserialize(cls, agent_specs_pb: PbAgentSpecs):
        """
        Factory function building an AgentSpecs instance from a AgentSpecs protobuf message
        """
        return cls(agent_specs_pb)



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import torch
from coltra import CAgent, DAgent, Observation
from coltra.models import BaseModel
from torch import nn
from torch.nn import functional as F

from cogment_lab.core import CogmentActor


class ColtraActor(CogmentActor):
    def __init__(self, model: BaseModel):
        super().__init__(model)
        self.model = model
        self.agent = DAgent(self.model) if self.model.discrete else CAgent(self.model)

    async def act(self, observation: np.ndarray, rendered_frame=None):
        obs = Observation(vector=observation)
        action, _, _ = self.agent.act(obs)
        return action.discrete


class NNActor(CogmentActor):
    def __init__(self, network: nn.Module, device: str = "cpu"):
        super().__init__(network=network)
        self.network = network
        self.device = device
        self.eps = 0.0
        self.num_actions: int | None = None
        self.rng = np.random.default_rng(0)

    async def act(self, observation: np.ndarray, rendered_frame=None) -> int:
        if self.num_actions is None:
            observation = observation.copy()
            obs = torch.from_numpy(observation).float().to(self.device)
            [act_probs] = self.network(obs)
            self.num_actions = act_probs.shape[0]

        if self.eps > 0.0 and self.rng.random() < self.eps:
            return self.rng.integers(0, self.num_actions)

        observation = observation.copy()
        obs = torch.from_numpy(observation).float().to(self.device)
        [act_probs] = self.network(obs)
        return act_probs.detach().cpu().numpy().argmax()

    def set_eps(self, eps: float):
        self.eps = eps


class BoltzmannActor(CogmentActor):
    def __init__(self, network: nn.Module, device: str = "cpu"):
        super().__init__(network=network)
        self.network = network
        self.device = device
        self.temperature = 1.0
        self.num_actions: int | None = None
        self.rng = np.random.default_rng(0)

    async def act(self, observation: np.ndarray, rendered_frame=None) -> int:
        observation = observation.copy()
        obs = torch.from_numpy(observation).float().to(self.device)
        with torch.no_grad():
            [act_vals] = self.network(obs)
            act_probs = F.softmax(act_vals / self.temperature, dim=0)

            action = int(torch.multinomial(act_probs, 1).item())

        return action

    def set_temperature(self, temperature: float):
        self.temperature = temperature



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import signal
from multiprocessing import Queue
from typing import Any

from cogment_lab.humans.actor import run_cogment_actor, start_fastapi
from cogment_lab.utils.runners import setup_logging


# def human_actor_runner(
#     app_port: int = 8000,
#     cogment_port: int = 8999,
#     log_file: str | None = None
# ):
#     """Runs the human actor along with the FastAPI server"""
#     if log_file:
#         setup_logging(log_file)
#
#     # Queues for communication between FastAPI and Cogment actor
#     app_to_actor = asyncio.Queue()
#     actor_to_app = asyncio.Queue()
#
#     # Asyncio tasks for the FastAPI server and Cogment actor
#     fastapi_task = start_fastapi(port=app_port, send_queue=app_to_actor, recv_queue=actor_to_app)
#     cogment_task = asyncio.create_task(run_cogment_actor(port=cogment_port, send_queue=actor_to_app, recv_queue=app_to_actor))
#
#     # Run the asyncio event loop
#     asyncio.run(asyncio.gather(fastapi_task, cogment_task))


async def shutdown():
    tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
    for task in tasks:
        task.cancel()
    await asyncio.gather(*tasks, return_exceptions=True)
    asyncio.get_event_loop().stop()


def signal_handler(sig, frame):
    asyncio.create_task(shutdown())


async def human_actor_main(
    app_port: int,
    cogment_port: int,
    signal_queue: Queue,
    actions: list[str] | None = None,
    fps: float = 30,
    html_override: str | None = None,
    file_override: str | None = None,
    jinja_parameters: dict[str, Any] | None = None,
):
    app_to_actor = asyncio.Queue()
    actor_to_app = asyncio.Queue()
    fastapi_task = asyncio.create_task(
        start_fastapi(
            port=app_port,
            send_queue=app_to_actor,
            recv_queue=actor_to_app,
            actions=actions,
            fps=fps,
            html_override=html_override,
            file_override=file_override,
            jinja_parameters=jinja_parameters,
        )
    )
    cogment_task = asyncio.create_task(
        run_cogment_actor(
            port=cogment_port,
            send_queue=actor_to_app,
            recv_queue=app_to_actor,
            signal_queue=signal_queue,
        )
    )

    await asyncio.gather(fastapi_task, cogment_task)


def human_actor_runner(
    app_port: int,
    cogment_port: int,
    signal_queue: Queue,
    log_file: str | None = None,
    actions: list[str] | None = None,
    fps: float = 30,
    html_override: str | None = None,
    file_override: str | None = None,
    jinja_parameters: dict[str, Any] | None = None,
):
    if log_file:
        setup_logging(log_file)

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    for sig in [signal.SIGINT, signal.SIGTERM]:
        loop.add_signal_handler(sig, lambda s=sig, frame=None: signal_handler(s, frame))

    try:
        loop.run_until_complete(
            human_actor_main(
                app_port,
                cogment_port,
                signal_queue,
                actions,
                fps,
                html_override,
                file_override,
                jinja_parameters,
            )
        )
    finally:
        loop.close()



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import atexit
import datetime
import logging
import multiprocessing as mp
import os
from asyncio import Task
from collections.abc import Callable, Coroutine, Sequence
from multiprocessing import Process, Queue
from typing import Any

import cogment
from cogment.control import Controller
from cogment.datastore import Datastore

from cogment_lab.actors.runner import actor_runner
from cogment_lab.core import BaseActor, BaseEnv
from cogment_lab.envs.runner import env_runner
from cogment_lab.generated import cog_settings, data_pb2
from cogment_lab.humans.runner import human_actor_runner
from cogment_lab.utils.trial_utils import (
    TrialData,
    format_data_multiagent,
    get_actor_params,
)


AgentName = str
ImplName = str
TrialName = str


class Cogment:
    """Main Cogment class for managing experiments"""

    controller: Controller
    datastore: Datastore

    def __init__(
        self,
        user_id: str = "cogment_lab",
        torch_mode: bool = False,
        log_dir: str | None = None,
        mp_method: str = "spawn",
        orchestrator_port: int = 9000,
        datastore_port: int = 9003,
    ):
        """Initializes the Cogment instance

        Args:
            user_id (str, optional): User ID. Defaults to "cogment_lab".
            torch_mode (bool, optional): Whether to use PyTorch multiprocessing. Defaults to False.
            log_dir (str, optional): Directory to store logs. Defaults to "logs".
            mp_method (str, optional): Multiprocessing method to use. Defaults to "spawn". WARNING: Methods other than spawn are likely to not work. Use at your own risk.
        """
        try:
            mp.set_start_method(mp_method)
        except RuntimeError:
            pass
        self.processes: dict[ImplName, Process] = {}
        self.tasks: dict[ImplName, Task] = {}

        self._register_shutdown_hook()
        self.torch_mode = torch_mode
        self.log_dir = log_dir

        self.envs: dict[ImplName, BaseEnv] = {}
        self.actors: dict[ImplName, BaseActor] = {}

        self.orchestrator_endpoint = f"grpc://localhost:{orchestrator_port}"
        self.datastore_endpoint = f"grpc://localhost:{datastore_port}"

        self.context = cogment.Context(cog_settings=cog_settings, user_id=user_id)
        controller = self.context.get_controller(endpoint=cogment.Endpoint(self.orchestrator_endpoint))
        datastore = self.context.get_datastore(endpoint=cogment.Endpoint(self.datastore_endpoint))

        assert isinstance(
            controller, Controller
        ), "self.controller is not an instance of Controller. Please report this."
        assert isinstance(datastore, Datastore), "self.datastore is not an instance of Datastore. Please report this."

        self.controller = controller
        self.datastore = datastore

        self.env_ports: dict[ImplName, int] = {}
        self.actor_ports: dict[ImplName, int] = {}

        self.mp_ctx = mp.get_context(mp_method) if mp_method else mp.get_context()

        self.trial_envs: dict[TrialName, ImplName] = {}

    def _add_process(
        self,
        target: Callable,
        args: tuple,
        name: ImplName,
        use_torch: bool | None = None,
        force: bool = False,
    ):
        """Adds a process to the list of processes

        Args:
            target (Callable): The process target function
            args (tuple): Arguments for the process target
            name (ImplName): Name of the process
            use_torch (bool | None, optional): Whether to use PyTorch multiprocessing. Defaults to None.
            force (bool, optional): Whether to force adding the process if it already exists. Defaults to False.

        Raises:
            ValueError: If the process already exists and force is False
        """
        if use_torch is None:
            use_torch = self.torch_mode

        if name in self.processes and not force:
            raise ValueError(f"Process {name} already exists")

        if use_torch:
            from torch.multiprocessing import Process as TorchProcess

            p = TorchProcess(target=target, args=args)
        else:
            p = self.mp_ctx.Process(target=target, args=args, name=f"cogment_lab_{name}")  # type: ignore
        p.start()
        self.processes[name] = p

    def _add_task(self, target: Coroutine, name: ImplName) -> Task:
        """Adds a task to the list of tasks

        Args:
            target (Coroutine): The task target function
            name (ImplName): Name of the task

        Returns:
            Task: The task instance
        """
        if name in self.tasks:
            raise ValueError(f"Task {name} already exists")

        task = asyncio.create_task(target)
        self.tasks[name] = task
        return task

    def run_env(
        self,
        env: BaseEnv,
        env_name: ImplName,
        port: int = 9001,
        log_file: str | None = None,
    ) -> Coroutine[None, None, bool]:
        """Given an environment, runs it in a subprocess

        Args:
            env (BaseEnv): The environment instance
            env_name (ImplName): Name for the environment
            port (int, optional): Port to run the environment on. Defaults to 9001.
            log_file (str | None, optional): Log file path. Defaults to None.

        Returns:
            bool: Whether the environment startup succeeded
        """
        env_class = type(env)
        env_args = env.args
        env_kwargs = env.kwargs

        signal_queue = Queue(1)

        if self.log_dir is not None and log_file:
            log_file = os.path.join(self.log_dir, log_file)

        self._add_process(
            target=env_runner,
            name=env_name,
            args=(
                env_class,
                env_args,
                env_kwargs,
                env_name,
                signal_queue,
                port,
                log_file,
            ),
        )
        logging.info(f"Started environment {env_name} on port {port} with log file {log_file}")

        self.envs[env_name] = env
        self.env_ports[env_name] = port

        return self.is_ready(signal_queue)

    def run_actor(
        self,
        actor: BaseActor,
        actor_name: ImplName,
        port: int = 9002,
        log_file: str | None = None,
    ) -> Coroutine[None, None, bool]:
        """Given an actor, runs it

        Args:
            actor (BaseActor): The actor instance
            actor_name (ImplName): Name for the actor
            port (int, optional): Port to run the actor on. Defaults to 9002.
            log_file (str | None, optional): Log file path. Defaults to None.

        Returns:
            bool: Whether the actor startup succeeded
        """
        actor_class = type(actor)
        actor_args = actor.args
        actor_kwargs = actor.kwargs

        signal_queue = Queue(1)

        if self.log_dir is not None and log_file:
            log_file = os.path.join(self.log_dir, log_file)

        self._add_process(
            target=actor_runner,
            name=actor_name,
            args=(
                actor_class,
                actor_args,
                actor_kwargs,
                actor_name,
                signal_queue,
                port,
                log_file,
            ),
        )
        logging.info(f"Started actor {actor_name} on port {port} with log file {log_file}")

        self.actors[actor_name] = actor
        self.actor_ports[actor_name] = port

        return self.is_ready(signal_queue)

    def run_local_actor(
        self,
        actor: BaseActor,
        actor_name: ImplName,
        port: int = 9002,
        log_file: str | None = None,
    ) -> Task:
        """Given an actor, runs it locally

        Args:
            actor (BaseActor): The actor instance
            actor_name (ImplName): Name for the actor
            port (int, optional): Port to run the actor on. Defaults to 9002.
            log_file (str | None, optional): Log file path. Defaults to None.

        Returns:
            bool: Whether the actor startup succeeded
        """

        if self.log_dir is not None and log_file:
            log_file = os.path.join(self.log_dir, log_file)

        self.context.register_actor(impl=actor.impl, impl_name=actor_name, actor_classes=["player"])

        serve = self._add_task(
            self.context.serve_all_registered(cogment.ServedEndpoint(port=port)),
            actor_name,
        )

        logging.info(f"Started actor {actor_name} on port {port} with log file {log_file}")

        self.actors[actor_name] = actor
        self.actor_ports[actor_name] = port

        return serve

    def run_web_ui(
        self,
        app_port: int = 8000,
        cogment_port: int = 8999,
        actions: list[str] | dict[str, Any] = [],
        log_file: str | None = None,
        fps: int = 30,
        html_override: str | None = None,
        file_override: str | None = None,
        jinja_parameters: dict[str, Any] | None = None,
    ) -> Coroutine[None, None, bool]:
        """Runs the human actor in a separate process

        Args:
            app_port (int, optional): Port for web UI. Defaults to 8000.
            cogment_port (int, optional): Port for Cogment connection. Defaults to 8999.
            actions (list[str], optional): Allowed actions. Defaults to [].
            log_file (str | None, optional): Log file path. Defaults to None.
            fps (int, optional): Frames per second for environment. Defaults to 30.
            html_override (str | None, optional): HTML override file path. Defaults to None.
            file_override (str | None, optional): File override file path. Defaults to None.
            jinja_parameters (dict[str, Any] | None, optional): Jinja parameters for the HTML override. Defaults to None.

        Returns:
            bool: Whether the web UI startup succeeded
        """

        signal_queue = Queue(1)

        if self.log_dir is not None and log_file:
            log_file = os.path.join(self.log_dir, log_file)

        self._add_process(
            target=human_actor_runner,
            name="web_ui",
            args=(
                app_port,
                cogment_port,
                signal_queue,
                log_file,
                actions,
                fps,
                html_override,
                file_override,
                jinja_parameters,
            ),
        )
        logging.info(f"Started web UI on port {app_port} with log file {log_file}")

        self.actor_ports["web_ui"] = cogment_port

        return self.is_ready(signal_queue)

    def stop_service(self, name: ImplName, timeout: float = 1.0):
        """Stops a process or a task.

        Args:
            name (str): Name of the process or task to stop
            timeout (float, optional): How long (in seconds) to wait for the process to stop before killing it. Defaults to 1.0.
        """
        if name in self.processes:
            self._stop_process(name, timeout)
        elif name in self.tasks:
            self._stop_task(name)
        else:
            raise ValueError(f"Service {name} does not exist")

    def _stop_process(self, name: ImplName, timeout: float = 1.0):
        """Stops a process

        Args:
            name (str): Name of the process to stop
            timeout (float, optional): How long (in seconds) to wait for the process to stop before killing it. Defaults to 1.0.
        """
        if name not in self.processes:
            raise ValueError(f"Process {name} does not exist")
        logging.info(f"Stopping process {name}")
        process = self.processes[name]
        if timeout == 0.0:
            process.kill()
            process.join()
        else:
            process.terminate()
            process.join(timeout=timeout)
            if process.is_alive():
                process.kill()
                process.join()
        del self.processes[name]

    def _stop_task(self, name: ImplName):
        """Stops a task

        Args:
            name (str): Name of the task to stop
        """
        if name not in self.tasks:
            raise ValueError(f"Task {name} does not exist")
        logging.info(f"Stopping task {name}")
        task = self.tasks[name]
        task.cancel()
        del self.tasks[name]

    def stop_all_services(self, timeout: float = 1.0):
        """Stops all processes and tasks

        Args:
            timeout (float, optional): How long (in seconds) to wait for the processes to stop before killing them. Defaults to 1.0.
        """
        for name in list(self.processes.keys()):
            self._stop_process(name, timeout)
        for name in list(self.tasks.keys()):
            self._stop_task(name)

    def _cleanup_wrapper(self):
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        asyncio.run(self.cleanup())
        loop.close()

    async def cleanup(self, timeout: float = 1.0):
        """Cleans up all processes. Idempotent."""
        tasks = list(self.tasks.keys())
        processes = list(self.processes.keys())

        for name in tasks:
            try:
                self._stop_task(name)
            except Exception as e:
                logging.warning(f"Failed to stop task {name}: {e}")

        for name in processes:
            try:
                self._stop_process(name, timeout)
            except Exception as e:
                logging.warning(f"Failed to stop process {name}: {e}")

        try:
            await self.context._grpc_server.stop(None)  # type: ignore
        except asyncio.exceptions.CancelledError:
            logging.info("Server already stopped")
        except AttributeError:
            logging.info("Server not started")

    def _register_shutdown_hook(self):
        """Registers the cleanup method to run on shutdown"""
        atexit.register(self._cleanup_wrapper)

    async def start_trial(
        self,
        env_name: ImplName,
        actor_impls: dict[AgentName, ImplName] | ImplName,
        session_config: dict[str, Any] = {},
        trial_name: TrialName | None = None,
    ):
        """Starts a new trial

        Args:
            env_name (ImplName): Name of the environment implementation
            actor_impls (dict[AgentName, ImplName] | ImplName): Actor implementations mapped to agent names
            session_config (dict[str, Any]): kwargs for the environment session
            trial_name (str | None, optional): Trial name. Defaults to None.

        Returns:
            str: The trial ID
        """
        if trial_name is None:
            trial_name = f"{env_name}-{datetime.datetime.now().isoformat()}"

        if isinstance(actor_impls, str):
            actor_impls = {"gym": actor_impls}

        env = self.envs[env_name]
        actor_params = [
            get_actor_params(
                name=agent_name,
                implementation=actor_impl,
                agent_specs=env.agent_specs[agent_name],
                endpoint=f"grpc://localhost:{self.actor_ports[actor_impl]}",
            )
            for agent_name, actor_impl in actor_impls.items()
        ]

        env_config = data_pb2.EnvironmentConfig(**session_config)  # type: ignore

        trial_params = cogment.TrialParameters(
            cog_settings,
            environment_name=env_name,
            environment_endpoint=f"grpc://localhost:{self.env_ports[env_name]}",
            environment_config=env_config,
            actors=actor_params,
            environment_implementation=env_name,
            datalog_endpoint=self.datastore_endpoint,
        )

        trial_id = await self.controller.start_trial(trial_id_requested=trial_name, trial_params=trial_params)

        logging.info(f"Started trial {trial_id} with name {trial_name}")

        self.trial_envs[trial_id] = env_name

        return trial_id

    async def get_trial_data(
        self,
        trial_id: str,
        env_name: str | None = None,
        fields: Sequence[str] = (
            "observations",
            "actions",
            "rewards",
            "done",
            "next_observations",
            "last_observation",
        ),
        use_tqdm: bool = False,
        tqdm_kwargs: dict[str, Any] | None = None,
    ) -> dict[str, TrialData]:
        """Gets trial data from the datastore, formatting it appropriately."""
        if env_name is None:
            env_name = self.trial_envs[trial_id]
        env = self.envs[env_name]
        agent_specs = env.agent_specs

        data = await format_data_multiagent(
            datastore=self.datastore,
            trial_id=trial_id,
            actor_agent_specs=agent_specs,
            fields=fields,
            use_tqdm=use_tqdm,
            tqdm_kwargs=tqdm_kwargs,
        )

        return data

    async def get_trial(self, trial_id: str):
        """Gets a trial by ID

        Args:
            trial_id (str): The trial ID

        Returns:
            Trial: The trial instance
        """
        [trial] = await self.datastore.get_trials(ids=[trial_id])
        return trial

    def __del__(self):
        """Cleanup on delete"""
        self.stop_all_services()

    async def is_ready(self, queue: Queue):
        """Waits for a readiness signal on a queue

        Args:
            queue (Queue): The queue to wait on

        Returns:
            Any: The object that was put on the queue
        """
        while queue.empty():
            await asyncio.sleep(0.1)
        return queue.get()



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import Any

import cogment
import gymnasium as gym
import numpy as np
from cogment import ActorParameters
from cogment.datastore import Datastore, DatastoreSample
from tqdm.auto import tqdm

from cogment_lab.generated import cog_settings
from cogment_lab.specs import AgentSpecs
from cogment_lab.utils.grpc import extend_actor_config


def get_actor_params(
    name: str,
    implementation: str,
    agent_specs: AgentSpecs,
    endpoint: str = "grpc://localhost:9002",
    base_params: dict | None = None,
    run_id: str = "run_id",
    seed: int = 0,
) -> ActorParameters:
    """
    Create and return actor parameters for a given actor.

    Args:
        name (str): The name of the actor.
        implementation (str): The implementation type of the actor.
        agent_specs (AgentSpecs): The agent specifications.
        endpoint (str): The endpoint URL, defaults to "grpc://localhost:9002".
        base_params (dict | None): Base parameters for the actor, optional.
        run_id (str): The run ID, defaults to "run_id".
        seed (int): The seed for randomness, defaults to 0.

    Returns:
        ActorParameters: The configured actor parameters.
    """
    if base_params is None:
        base_params = {}
    params = ActorParameters(
        cog_settings,
        name=name,
        class_name="player",
        endpoint=endpoint,
        implementation=implementation,
        config=extend_actor_config(base_params, run_id, agent_specs.serialize(), seed),
    )
    return params


@dataclass
class TrialData:
    """
    Dataclass to store structured trial data for reinforcement learning.

    Attributes:
        observations (np.ndarray | dict[str, np.ndarray] | None): Observations from the trial.
        actions (np.ndarray | dict[str, np.ndarray] | None): Actions taken during the trial.
        rewards (np.ndarray | None): Rewards received during the trial.
        done (np.ndarray | None): Done flags for each step of the trial.
        next_observations (np.ndarray | dict[str, np.ndarray] | None): Next observations after each step.
        last_observation (np.ndarray | dict[str, np.ndarray] | None): The final observation of the trial.
    """

    observations: np.ndarray | dict[str, np.ndarray] | None = field(default=None)
    actions: np.ndarray | dict[str, np.ndarray] | None = field(default=None)
    rewards: np.ndarray | None = field(default=None)
    done: np.ndarray | None = field(default=None)
    next_observations: np.ndarray | dict[str, np.ndarray] | None = field(default=None)
    last_observation: np.ndarray | dict[str, np.ndarray] | None = field(default=None)


def initialize_buffer(space: gym.Space | None, length: int) -> np.ndarray | dict[str, np.ndarray]:
    """
    Initializes a buffer based on the given gym space and length.

    Args:
        space (gym.Space | None): The gym space to base the buffer on. If None, an empty buffer is created.
        length (int): The length of the buffer.

    Returns:
        Union[np.ndarray, Dict[str, np.ndarray]]: The initialized buffer, either as an ndarray or a dictionary of ndarrays.
    """
    if space is None:
        return np.empty((length,), dtype=np.float32)
    elif isinstance(space, gym.spaces.Dict):
        return {key: np.empty((length,) + space[key].shape, dtype=space[key].dtype) for key in space.spaces.keys()}  # type: ignore
    elif isinstance(space, gym.spaces.Tuple):
        return {i: np.empty((length,) + space[i].shape, dtype=space[i].dtype) for i in range(len(space.spaces))}  # type: ignore
    else:  # Simple space
        assert isinstance(space, gym.spaces.Space)
        assert space.shape is not None
        return np.empty((length,) + space.shape, dtype=space.dtype)


def write_to_buffer(
    buffer: np.ndarray | dict[str, np.ndarray],
    data: np.ndarray | dict[str, Any] | bool,
    idx: int,
):
    """
    Writes data to a specified index in the buffer.

    Args:
        buffer (np.ndarray | dict[str, np.ndarray]): The buffer to write data to.
        data (np.ndarray | dict[str, Any]): The data to write.
        idx (int): The index at which to write the data.

    """
    if isinstance(buffer, dict):
        for key in buffer.keys():
            buffer[key][idx] = data[key]  # type: ignore
    else:
        buffer[idx] = data


def extract_data_from_samples(
    samples: list[DatastoreSample],
    agent_specs: AgentSpecs,
    fields: Sequence[str] = (
        "observations",
        "actions",
        "rewards",
        "done",
        "next_observations",
        "last_observation",
    ),
    actor_name: str = "gym",
) -> TrialData:
    """
    Extracts trial data into a TrialData instance from a list of DatastoreSamples.

    Args:
        samples (list[DatastoreSample]): The samples to extract data from.
        fields (Sequence[str]): The fields to extract into the TrialData.
        agent_specs (AgentSpecs): The agent specifications.
        actor_name (str): The name of the actor to extract data for.

    Returns:
       TrialData: The extracted trial data.
    """
    sample_count = len(samples)
    if sample_count == 0:
        raise ValueError("No samples provided")

    cog_observation_space = agent_specs.get_observation_space()
    observation_space = cog_observation_space.gym_space

    cog_action_space = agent_specs.get_action_space()
    action_space = cog_action_space.gym_space

    data = TrialData()
    if "observations" in fields:
        data.observations = initialize_buffer(observation_space, sample_count - 1)  # type: ignore
    if "actions" in fields:
        data.actions = initialize_buffer(action_space, sample_count - 1)  # type: ignore
    if "rewards" in fields:
        data.rewards = initialize_buffer(None, sample_count - 1)  # type: ignore
    if "done" in fields:
        data.done = initialize_buffer(None, sample_count - 1)  # type: ignore
        data.done[-1] = True  # type: ignore
    if "next_observations" in fields:
        data.next_observations = initialize_buffer(observation_space, sample_count - 1)
    if "last_observation" in fields:
        data.last_observation = initialize_buffer(observation_space, 1)

    for i, sample in enumerate(samples[:-1]):
        if "observations" in fields:
            obs = cog_observation_space.deserialize(sample.actors_data[actor_name].observation).value
            write_to_buffer(data.observations, obs, i)  # type: ignore
        if "actions" in fields:
            action = cog_action_space.deserialize(sample.actors_data[actor_name].action).value
            write_to_buffer(data.actions, action, i)  # type: ignore
        if "rewards" in fields:
            write_to_buffer(data.rewards, sample.actors_data[actor_name].reward, i)  # type: ignore
        if "done" in fields and i < sample_count - 2:
            write_to_buffer(data.done, False, i)  # type: ignore
        if "next_observations" in fields:
            next_obs = cog_observation_space.deserialize(samples[i + 1].actors_data[actor_name].observation).value
            write_to_buffer(data.next_observations, next_obs, i)  # type: ignore
    if "last_observation" in fields:
        last_obs = (
            agent_specs.get_observation_space().deserialize(samples[-1].actors_data[actor_name].observation).value
        )
        write_to_buffer(data.last_observation, last_obs, 0)  # type: ignore

    return data


def extract_rewards_from_samples(
    samples: list[DatastoreSample],
    actor_name: str = "gym",
) -> np.ndarray:
    """
    Extracts rewards from trial samples into a TrialData instance.

    Args:
        samples (list[DatastoreSample]): The samples to extract rewards from.
        actor_name (str): The name of the actor to extract rewards for.

    Returns:
        TrialData: The extracted rewards.
    """
    sample_count = len(samples)

    rewards = initialize_buffer(None, sample_count - 1)
    assert isinstance(rewards, np.ndarray)

    for i, sample in enumerate(samples[:-1]):
        write_to_buffer(rewards, sample.actors_data[actor_name].reward, i)

    return rewards


def concat_trial_field(
    field_data: list[np.ndarray | dict[str, np.ndarray] | None],
) -> np.ndarray | dict[str, np.ndarray] | None:
    """
    Concatenates a list of fields (either np.ndarray or dict of np.ndarrays) from TrialData instances.
    Filters out None values before concatenation.

    Args:
    field_data: List of fields to be concatenated.

    Returns:
    Concatenated field as np.ndarray or dict of np.ndarrays, or None if all elements are None.
    """
    # Filter out None values
    valid_field_data = [data for data in field_data if data is not None]

    if not valid_field_data:
        return None

    if all(isinstance(data, np.ndarray) for data in valid_field_data):
        return np.concatenate(valid_field_data, axis=0)  # type: ignore
    elif all(isinstance(data, dict) for data in valid_field_data):
        keys = valid_field_data[0].keys()  # type: ignore
        return {key: np.concatenate([data[key] for data in valid_field_data], axis=0) for key in keys}
    else:
        raise TypeError("Inconsistent field types in TrialData list.")


def concatenate(trial_data_list: list[TrialData]) -> TrialData:
    """
    Concatenates a list of TrialData instances into a single TrialData instance.

    Args:
    trial_data_list: List of TrialData instances to be concatenated.

    Returns:
    A single concatenated TrialData instance.
    """
    observations = concat_trial_field([trial.observations for trial in trial_data_list])
    actions = concat_trial_field([trial.actions for trial in trial_data_list])
    rewards = concat_trial_field([trial.rewards for trial in trial_data_list])
    done = concat_trial_field([trial.done for trial in trial_data_list])
    next_observations = concat_trial_field([trial.next_observations for trial in trial_data_list])

    # Handle 'last_observation' separately
    last_trial = trial_data_list[-1]
    last_observation = (
        last_trial.last_observation if last_trial.last_observation is not None else last_trial.next_observations[-1]  # type: ignore
    )

    return TrialData(
        observations=observations,
        actions=actions,
        rewards=rewards,  # type: ignore
        done=done,  # type: ignore
        next_observations=next_observations,
        last_observation=last_observation,
    )


async def format_data_multiagent(
    datastore: Datastore,
    trial_id: str,
    actor_agent_specs: dict[str, AgentSpecs],
    fields: Sequence[str] = (
        "observations",
        "actions",
        "rewards",
        "done",
        "next_observations",
        "last_observation",
    ),
    use_tqdm: bool = False,
    tqdm_kwargs: dict[str, Any] | None = None,
) -> dict[str, TrialData]:
    """
    Formats trial data from a multiagent Cogment trial into structured formats for reinforcement learning.

    Args:
        datastore (Datastore): The datastore to fetch trial data from.
        trial_id (str): The identifier of the trial.
        actor_agent_specs (dict[str, EnvironmentSpecs]): A dictionary mapping actor IDs to their environment specifications.
        fields (List[str]): The list of fields to include in the formatted data.
        tqdm_kwargs (dict[str, Any] | None): Optional keyword arguments to pass to tqdm.

    Returns:
        dict[str, TrialData]: A dictionary mapping actor IDs to their formatted trial data.
    """
    if tqdm_kwargs is None:
        tqdm_kwargs = {}

    trials = []
    while len(trials) == 0:
        try:
            trials = await datastore.get_trials([trial_id])
        except cogment.CogmentError:
            continue

    # Initialize a dictionary to store samples for each actor
    actor_samples = {actor_id: [] for actor_id in actor_agent_specs.keys()}
    actor_reward_samples = {actor_id: [] for actor_id in actor_agent_specs.keys()}

    # Get all samples
    async for sample in tqdm(datastore.all_samples(trials), disable=not use_tqdm, **tqdm_kwargs):  # type: ignore
        for actor_id in actor_agent_specs.keys():
            # Add the sample to the list for an actor if the observation for that actor is not None
            if (
                sample.actors_data.get(actor_id)
                and sample.actors_data[actor_id].observation is not None
                and sample.actors_data[actor_id].observation.value is not None
                and sample.actors_data[actor_id].observation.value.raw_data != b""
            ):
                actor_samples[actor_id].append(sample)

            if (
                sample.actors_data.get(actor_id)
                and sample.actors_data[actor_id].reward is not None
                and sample.actors_data[actor_id].reward == sample.actors_data[actor_id].reward  # Check for NaN
            ):
                actor_reward_samples[actor_id].append(sample)

    # Extract data for each actor
    actor_data = {}

    for actor_id, samples in actor_samples.items():
        actor_data[actor_id] = extract_data_from_samples(
            samples, actor_agent_specs[actor_id], fields, actor_name=actor_id
        )

    for actor_id, reward_samples in actor_reward_samples.items():
        actor_data[actor_id].rewards = extract_rewards_from_samples(reward_samples, actor_name=actor_id)

    return actor_data



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

from cogment.actor import ActorSession
from cogment.environment import EnvironmentSession
from cogment.model_registry_v2 import ModelRegistry
from cogment.session import ActorInfo, RecvEvent

from cogment_lab.specs import AgentSpecs
from cogment_lab.specs.action_space import Action, ActionSpace
from cogment_lab.specs.observation_space import Observation, ObservationSpace


class ActorSessionHelper:
    """
    Cogment Verse actor session helper

    Provides additional methods to the regular Cogment actor session.
    """

    def __init__(self, actor_session: ActorSession, model_registry: ModelRegistry | None):
        self.actor_session = actor_session
        self.agent_specs = AgentSpecs.deserialize(self.actor_session.config.agent_specs)
        self.action_space = self.agent_specs.get_action_space(seed=self.actor_session.config.seed)
        self.observation_space = self.agent_specs.get_observation_space()
        self.model_registry = model_registry

    def get_action_space(self) -> ActionSpace:
        return self.action_space

    def get_observation_space(self) -> ObservationSpace:
        return self.observation_space

    def get_observation(self, event: RecvEvent) -> Observation | None:
        """
        Return the cogment verse observation for the current event.

        If the event does not contain an observation, return None.
        """
        if not event.observation:
            return None

        return self.observation_space.deserialize(event.observation.observation)

    def get_render(self, event: RecvEvent) -> bytes | None:
        """
        Return the render for the current event.

        If the event does not contain a render, return None.
        """
        if not event.observation:
            return None

        return event.observation.render


class EnvironmentSessionHelper:
    """
    A session helper for environments.
    """

    actor_infos: list[ActorInfo]

    def __init__(
        self,
        environment_session: EnvironmentSession,
        agent_specs: dict[str, AgentSpecs],
    ):
        self.actor_infos = environment_session.get_active_actors()

        assert set(agent_specs.keys()) == {
            actor_info.actor_name for actor_info in self.actor_infos
        }, f"Agent specs and active actors do not match. {agent_specs.keys()} != {self.actor_infos}"

        # Mapping actor_name to actor_idx
        self.actor_idxs = {actor_info.actor_name: actor_idx for (actor_idx, actor_info) in enumerate(self.actor_infos)}
        # Mapping actor_idx to actor_info
        self.actors = [actor_info.actor_name for actor_info in self.actor_infos]

        if isinstance(agent_specs, AgentSpecs):
            agent_specs = {actor_name: agent_specs for actor_name in self.actors}

        self.agent_specs = agent_specs
        self.observation_spaces = {
            agent_id: specs.get_observation_space(environment_session.config.render_width)
            for (agent_id, specs) in agent_specs.items()
        }

    def get_observation_space(self, actor_name: str) -> ObservationSpace:
        return self.observation_spaces[actor_name]

    def get_action_space(self, actor_name: str) -> ActionSpace:
        return self.agent_specs[actor_name].get_action_space()

    def _get_actor_idx(self, actor_name: str) -> int:
        actor_idx = self.actor_idxs.get(actor_name)

        if actor_idx is None:
            raise RuntimeError(f"No actor with name [{actor_name}] found!")

        return actor_idx

    def get_action(self, tick_data: Any, actor_name: str) -> Action | None:
        # For environments, tick_datas are events
        event: RecvEvent = tick_data

        if not event.actions or not event.actions:
            return None

        actor_idx = self._get_actor_idx(actor_name)
        action_space = self.get_action_space(actor_name)

        return action_space.deserialize(
            event.actions[actor_idx].action,
        )

    def get_observation(self, tick_data: Any, actor_name: str):
        """
        Return the cogment verse observation of a given actor at a tick.

        If no observation, returns None.
        """
        raise NotImplementedError

    def get_player_actions(self, tick_data: Any, actor_name: str) -> Action | None:
        """
        Return the cogment verse player action of a given actor at a tick.

        If only a single player actor is present, no `actor_name` is required.

        If no action, returns None.
        """
        event = tick_data
        if not event.actions:
            return None

        actions = [
            self.get_action(actor_name, tick_data)
            for player_actor_name in self.actors
            if player_actor_name == actor_name
        ]
        if len(actions) == 0:
            raise RuntimeError(f"No player actors having name [{actor_name}]")
        return actions[0]

    def get_player_observations(self, tick_data: Any, actor_name: str):
        if actor_name is None:
            observations = [self.get_observation(tick_data, actor_name) for player_actor_name in self.actors]
            if len(observations) == 0:
                raise RuntimeError("No player actors")
            if len(observations) > 1:
                raise RuntimeError("More than 1 player actor, please provide an actor name")
            return observations[0]

        observations = [
            self.get_observation(tick_data, actor_name)
            for player_actor_name in self.actors
            if player_actor_name == actor_name
        ]
        if len(observations) == 0:
            raise RuntimeError(f"No player actors having name [{actor_name}]")
        return observations[0]



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import gymnasium as gym
import numpy as np

from cogment_lab.core import CogmentActor


class RandomActor(CogmentActor):
    def __init__(self, action_space: gym.spaces.Space):
        super().__init__(action_space)
        self.gym_action_space = action_space

    async def act(self, observation: Any, rendered_frame=None):
        return self.gym_action_space.sample()


class ConstantActor(CogmentActor):
    def __init__(self, action: Any):
        super().__init__(action)
        if isinstance(action, list):
            action = np.array(action)
        self.action = action

    async def act(self, observation: Any, rendered_frame=None):
        return self.action



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import importlib
import logging
import os
from typing import Any, Callable

import gymnasium as gym
from cogment.environment import EnvironmentSession

from cogment_lab.core import CogmentEnv, State
from cogment_lab.session_helpers import EnvironmentSessionHelper
from cogment_lab.specs import AgentSpecs


log = logging.getLogger(__name__)

# configure pygame to use a dummy video server to be able to render headlessly
os.environ["SDL_VIDEODRIVER"] = "dummy"


class GymEnvironment(CogmentEnv):
    """
    Gymnasium integration for Cogment.

    Exposes a Gymnasium environment as a Cogment environment.
    """

    session_helper: EnvironmentSessionHelper
    actor_name = "gym"

    def __init__(
        self,
        env_id: str | Callable[..., gym.Env],
        registration: str | None = None,
        make_kwargs: dict[str, Any] | None = None,
        reset_options: dict[str, Any] | None = None,
        render: bool = False,
        reinitialize: bool = False,
        dry: bool = False,
        sub_dry: bool = True,
    ):
        """
        Initialize the GymEnvironment.

        Args:
            env_id: The Gym environment ID.
            registration: Optional Gym registration string.
            make_kwargs: Optional args to pass to gym.make().
            reset_options: Optional reset options to pass to env.reset().
            render: Whether to render the environment.
            reinitialize: Whether to reinitialize the environment each session.
            dry: Whether to abstain from initializing the environment in this process.
            sub_dry: Whether to abstain from initializing the environment in the initializer of the subprocess.
        """
        super().__init__(
            env_id=env_id,
            registration=registration,
            make_kwargs=make_kwargs,
            reset_options=reset_options,
            render=render,
            reinitialize=reinitialize,
            dry=sub_dry,
            sub_dry=sub_dry,
        )

        self.env_id = env_id
        self.registration = registration
        self.make_kwargs = make_kwargs or {}
        self.reset_options = reset_options or {}
        self.render = render
        self.reinitialize = reinitialize
        self.dry = dry

        if "render_mode" in self.make_kwargs:
            raise ValueError("render_mode cannot be set in make_kwargs")

        if self.render:
            self.make_kwargs["render_mode"] = "rgb_array"

        if isinstance(self.env_id, Callable):
            self.env_maker = self.env_id
        elif isinstance(self.env_id, str):
            self.env_maker = lambda **kwargs: gym.make(self.env_id, **kwargs)  # type: ignore
        else:
            raise ValueError(f"env_id must be a string or a callable, got {self.env_id}")

        if self.registration:
            importlib.import_module(self.registration)

        if not self.dry:
            self.env = self.env_maker(**self.make_kwargs)

            self.agent_specs = {
                "gym": AgentSpecs.create_homogeneous(
                    observation_space=self.env.observation_space,
                    action_space=self.env.action_space,
                )
            }

            self.initialized = True
        else:
            self.env = None
            self.agent_specs = {}
            self.initialized = False

        self._is_closed = False

    def get_implementation_name(self):
        """
        Get the name of the Gym environment.

        Returns:
            The Gym environment ID.
        """
        return self.env_id

    def get_agent_specs(self):
        """
        Get the agent specs.

        Returns:
            The agent specs dict.
        """
        return self.agent_specs

    async def initialize(self, state: State, environment_session: EnvironmentSession):
        """
        Initialize the environment session.

        Args:
            state: The Cogment state.
            environment_session: The Cogment environment session.

        Returns:
            The updated state.
        """

        logging.info("Initializing environment session")

        if not self.initialized and not self.reinitialize:
            self.env = self.env_maker(**self.make_kwargs)
            self.agent_specs = {
                "gym": AgentSpecs.create_homogeneous(
                    observation_space=self.env.observation_space,
                    action_space=self.env.action_space,
                )
            }

            state.env = self.env
            state.agent_specs = self.agent_specs
        elif self.initialized and not self.reinitialize:
            state.env = self.env
            state.agent_specs = self.agent_specs
        elif self.reinitialize:
            state.env = self.env_maker(**self.make_kwargs)
            state.agent_specs = {
                "gym": AgentSpecs.create_homogeneous(
                    observation_space=state.env.observation_space,
                    action_space=state.env.action_space,
                )
            }

        state.environment_session = environment_session
        state.session_helper = EnvironmentSessionHelper(environment_session, state.agent_specs)
        state.session_cfg = state.environment_session.config
        state.actors = state.session_helper.actors
        state.actor_name = state.session_helper.actors[0]

        self.initialized = True

        return state

    async def reset(self, state: State):
        """
        Reset the environment.

        Args:
            state: The Cogment state.

        Returns:
            A tuple with the updated state and a dict of observations.
        """

        logging.info("Resetting environment")

        obs, _info = state.env.reset(seed=state.session_cfg.seed, options=state.session_cfg.reset_args)  # THIS

        state.observation_space = state.session_helper.get_observation_space(self.actor_name)
        frame = state.env.render() if state.session_cfg.render else None
        observation = state.observation_space.create_serialize(value=obs, rendered_frame=frame, active=True, alive=True)

        return state, {"*": observation}

    async def read_actions(self, state: State, event):
        """
        Read the agent action from the event.

        Args:
            state: The Cogment state.
            event: The event from Cogment.

        Returns:
            The agent action value.
        """
        player_action = state.session_helper.get_action(tick_data=event, actor_name=self.actor_name)
        return player_action.value

    async def step(self, state: State, action):
        """
        Step the environment.

        Args:
            state: The Cogment state.
            action: The agent action.

        Returns:
            A tuple with the updated state, observations dict, rewards dict,
            terminateds dict, truncateds dict, and info dict.
        """

        logging.info("Stepping environment")

        obs, reward, terminated, truncated, info = state.env.step(action)
        logging.info("Stepping the environment")

        observation = state.observation_space.create_serialize(
            value=obs,
            rendered_frame=state.env.render() if state.session_cfg.render else None,
            active=True,
            alive=True,
        )

        # observations = [("*", observation)]
        observations = {"*": observation}
        rewards = {self.actor_name: reward}
        terminateds = {self.actor_name: terminated}
        truncateds = {self.actor_name: truncated}

        return state, observations, rewards, terminateds, truncateds, info

    async def end(self, state: State):
        """
        End the environment session.

        Args:
            state: The Cogment state.

        Returns:
            The updated state.
        """

        logging.info("Ending environment")
        state.env.close()
        if self.env is not None and not self._is_closed:
            self.env.close()
        self._is_closed = True



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import argparse
import logging
import sys

from cogment_lab.cli.download_cogment import download_cogment
from cogment_lab.constants import COGMENT_LAB_HOME


TEAL = "\033[36m"
RESET = "\033[0m"

custom_format = f"%(asctime)s {TEAL}[%(levelname)s] [%(name)s]{RESET} %(message)s [thread:%(thread)d]"

formatter = logging.Formatter(custom_format, datefmt="%Y-%m-%dT%H:%M:%S%z")

handler = logging.StreamHandler()
handler.setFormatter(formatter)

logging.basicConfig(level=logging.INFO, handlers=[handler])

sys.path.insert(0, "..")


def install_cogment():
    install_dir = COGMENT_LAB_HOME
    try:
        cogment_path = download_cogment(install_dir, "2.19.1")
        logging.info(f"Cogment installed successfully in [{cogment_path}].")
    except Exception as e:
        logging.error(f"Installation failed: {e}")


def main():
    parser = argparse.ArgumentParser(description="Cogment Lab CLI")
    subparsers = parser.add_subparsers(dest="command")

    # launch subcommand
    parser_launch = subparsers.add_parser("launch")
    parser_launch.add_argument("file")

    # env subcommand
    parser_env = subparsers.add_parser("env")
    parser_env.add_argument("file")

    parser_actor = subparsers.add_parser("actor")
    parser_actor.add_argument("file")

    parser_install = subparsers.add_parser("install")
    parser_install.add_argument("path", nargs="?")

    args = parser.parse_args()

    if args.command == "install":
        install_cogment()
    elif args.command == "launch":
        from cogment_lab.cli import launch

        launch.launch_main(args.file)
    else:
        print("Invalid command. Use 'launch' or `install`.")


if __name__ == "__main__":
    main()



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import subprocess

from cogment_lab.constants import COGMENT_LAB_HOME


def launch_service(service_name: str):
    cogment_path = COGMENT_LAB_HOME / "cogment"
    try:
        process = subprocess.Popen([cogment_path, "services", service_name])
        logging.info(f"{service_name} launched successfully. PID: {process.pid}")
        return process
    except Exception as e:
        logging.error(f"Failed to launch {service_name}: {e}")
        return None


def launch_main(command: str):
    services = command.split()
    processes = []

    if "base" in services or "all" in services:
        if len(services) > 1:
            raise ValueError("Cannot combine 'base' or 'all' with other services")

    if "base" in services:
        services_to_run = ["orchestrator", "trial_datastore"]
    elif "all" in services:
        services_to_run = [
            "orchestrator",
            "trial_datastore",
            "model_registry",
            "directory",
            "web_proxy",
        ]
    else:
        services_to_run = services

    for service in services_to_run:
        process = launch_service(service)
        if process:
            processes.append(process)

    for process in processes:
        process.wait()

    return processes



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gymnasium as gym
import numpy as np

from cogment_lab.constants import DEFAULT_RENDERED_WIDTH
from cogment_lab.generated.data_pb2 import Observation as PbObservation  # type: ignore

from .encode_rendered_frame import decode_rendered_frame, encode_rendered_frame
from .ndarray_serialization import deserialize_ndarray, serialize_ndarray


# pylint: disable=attribute-defined-outside-init
class Observation:
    """
    Cogment Verse actor observation

    Properties:
        flat_value:
            The observation value, as a flat numpy array.
        value:
            The observation value, as a numpy array.
        active: optional
            Boolean indicating if the object is active.
        alive: optional
            Boolean indicating if the object is alive.
        rendered_frame: optional
            Environment's rendered frame as a numpy array of RGB pixels.
    """

    def __init__(
        self,
        gym_space: gym.Space,
        pb_observation=None,
        value=None,
        active=None,
        alive=None,
        rendered_frame=None,
    ):
        """
        Observation constructor.
        Shouldn't be called directly, prefer the factory function of ObservationSpace.
        """

        self._gym_space = gym_space

        if pb_observation is not None:
            assert value is None
            assert active is None
            assert alive is None
            assert rendered_frame is None
            self._pb_observation = pb_observation
            return

        self._value = value
        self._active = active
        self._alive = alive
        self._rendered_frame = rendered_frame

        self._pb_observation = PbObservation(
            active=active,
            alive=alive,
        )

    def _compute_flat_value(self):
        if hasattr(self, "_value"):
            return gym.spaces.flatten(self._gym_space, self._value)

        if not self._pb_observation.value != b"" or self._pb_observation.value is None:
            return None

        return deserialize_ndarray(self._pb_observation.value)

    @property
    def flat_value(self):
        if not hasattr(self, "_flat_value"):
            self._flat_value = self._compute_flat_value()
        return self._flat_value

    def _compute_value(self):
        return gym.spaces.unflatten(self._gym_space, self.flat_value) if self.flat_value is not None else None

    @property
    def value(self):
        if not hasattr(self, "_value"):
            self._value = self._compute_value()
        return self._value

    def _deserialize_rendered_frame(self):
        if self._pb_observation.rendered_frame == b"" or self._pb_observation.rendered_frame is None:
            return None
        return decode_rendered_frame(self._pb_observation.rendered_frame)

    @property
    def rendered_frame(self):
        if not hasattr(self, "_rendered_frame"):
            self._rendered_frame = self._deserialize_rendered_frame()
        return self._rendered_frame

    @property
    def active(self):
        return self._pb_observation.active if self._pb_observation.active != b"" else self._active

    @property
    def alive(self):
        return self._pb_observation.alive if self._pb_observation.alive != b"" else self._alive

    def __repr__(self):
        return f"Observation(value={self.value.shape if isinstance(self.value, np.ndarray) else self.value}, active={self.active}, alive={self.alive}, rendered_frame={self.rendered_frame.shape if self.rendered_frame is not None else 'None'})@{hex(id(self))}"  # type: ignore

    def __str__(self):
        return self.__repr__()


class ObservationSpace:
    """
    Cogment Verse observation space

    Properties:
        gym_space:
            Wrapped Gym space for the observation values
        render_width:
            Maximum width for the serialized rendered frame in observations
    """

    def __init__(self, space: gym.Space, render_width: int = DEFAULT_RENDERED_WIDTH):
        """
        ObservationSpace constructor.
        Shouldn't be called directly, prefer the factory function of EnvironmentSpecs.
        """
        if isinstance(space, gym.spaces.Dict) and ("action_mask" in space.spaces):
            # Check the observation space defines an action_mask "component" (like PettingZoo does)
            assert "observation" in space.spaces
            assert len(space.spaces) == 2

            self.gym_space = space.spaces["observation"]
            self.action_mask_gym_space = space.spaces["action_mask"]
        else:
            # "Standard" observation space, no action_mask
            self.gym_space = space
            self.action_mask_gym_space = None

        # Other configuration
        self.render_width = render_width

    def create(
        self,
        value=None,
        active=None,
        alive=None,
        rendered_frame=None,
    ) -> Observation:
        """
        Create an Observation
        """
        return Observation(
            self.gym_space,
            value=value,
            active=active,
            alive=alive,
            rendered_frame=rendered_frame,
        )

    def serialize(
        self,
        observation: Observation,
    ) -> PbObservation:
        """
        Serialize an Observation to an Observation protobuf message
        """

        serialized_value = None
        if observation.value is not None:
            flat_value = gym.spaces.flatten(self.gym_space, observation.value)
            assert isinstance(flat_value, np.ndarray)
            serialized_value = serialize_ndarray(flat_value)

        serialized_rendered_frame = None
        if observation.rendered_frame is not None:
            serialized_rendered_frame = encode_rendered_frame(
                rendered_frame=observation.rendered_frame, max_size=self.render_width
            )

        return PbObservation(
            value=serialized_value,
            active=observation.active,
            alive=observation.alive,
            rendered_frame=serialized_rendered_frame,
        )

    def deserialize(self, pb_observation: PbObservation) -> Observation:
        """
        Deserialize an Observation from an Observation protobuf message
        """

        return Observation(self.gym_space, pb_observation=pb_observation)

    def create_serialize(
        self,
        value=None,
        active=None,
        alive=None,
        rendered_frame=None,
    ) -> PbObservation:
        """
        Create a serialized Observation
        """
        return self.serialize(
            self.create(
                value=value,
                active=active,
                alive=alive,
                rendered_frame=rendered_frame,
            )
        )



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: spaces.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import cogment_lab.generated.ndarray_pb2 as ndarray__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0cspaces.proto\x12\x12\x63ogment_lab.spaces\x1a\rndarray.proto"$\n\x08\x44iscrete\x12\t\n\x01n\x18\x01 \x01(\x05\x12\r\n\x05start\x18\x02 \x01(\x05"Z\n\x03\x42ox\x12(\n\x03low\x18\x02 \x01(\x0b\x32\x1b.cogment_lab.nd_array.Array\x12)\n\x04high\x18\x03 \x01(\x0b\x32\x1b.cogment_lab.nd_array.Array"5\n\x0bMultiBinary\x12&\n\x01n\x18\x01 \x01(\x0b\x32\x1b.cogment_lab.nd_array.Array":\n\rMultiDiscrete\x12)\n\x04nvec\x18\x01 \x01(\x0b\x32\x1b.cogment_lab.nd_array.Array"|\n\x04\x44ict\x12\x31\n\x06spaces\x18\x01 \x03(\x0b\x32!.cogment_lab.spaces.Dict.SubSpace\x1a\x41\n\x08SubSpace\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05space\x18\x02 \x01(\x0b\x32\x19.cogment_lab.spaces.Space"\x89\x02\n\x05Space\x12\x30\n\x08\x64iscrete\x18\x01 \x01(\x0b\x32\x1c.cogment_lab.spaces.DiscreteH\x00\x12&\n\x03\x62ox\x18\x02 \x01(\x0b\x32\x17.cogment_lab.spaces.BoxH\x00\x12(\n\x04\x64ict\x18\x03 \x01(\x0b\x32\x18.cogment_lab.spaces.DictH\x00\x12\x37\n\x0cmulti_binary\x18\x04 \x01(\x0b\x32\x1f.cogment_lab.spaces.MultiBinaryH\x00\x12;\n\x0emulti_discrete\x18\x05 \x01(\x0b\x32!.cogment_lab.spaces.MultiDiscreteH\x00\x42\x06\n\x04kindb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "spaces_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _DISCRETE._serialized_start = 51
    _DISCRETE._serialized_end = 87
    _BOX._serialized_start = 89
    _BOX._serialized_end = 179
    _MULTIBINARY._serialized_start = 181
    _MULTIBINARY._serialized_end = 234
    _MULTIDISCRETE._serialized_start = 236
    _MULTIDISCRETE._serialized_end = 294
    _DICT._serialized_start = 296
    _DICT._serialized_end = 420
    _DICT_SUBSPACE._serialized_start = 355
    _DICT_SUBSPACE._serialized_end = 420
    _SPACE._serialized_start = 423
    _SPACE._serialized_end = 688
# @@protoc_insertion_point(module_scope)



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ndarray.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rndarray.proto\x12\x14\x63ogment_lab.nd_array"\xb8\x01\n\x05\x41rray\x12*\n\x05\x64type\x18\x01 \x01(\x0e\x32\x1b.cogment_lab.nd_array.DType\x12\r\n\x05shape\x18\x02 \x03(\r\x12\x10\n\x08raw_data\x18\x03 \x01(\x0c\x12\x10\n\x08npy_data\x18\x04 \x01(\x0c\x12\x13\n\x0b\x64ouble_data\x18\x05 \x03(\x01\x12\x12\n\nint32_data\x18\x06 \x03(\x11\x12\x12\n\nint64_data\x18\x07 \x03(\x12\x12\x13\n\x0buint32_data\x18\x08 \x03(\r*\x83\x01\n\x05\x44Type\x12\x11\n\rDTYPE_UNKNOWN\x10\x00\x12\x11\n\rDTYPE_FLOAT32\x10\x01\x12\x11\n\rDTYPE_FLOAT64\x10\x02\x12\x0e\n\nDTYPE_INT8\x10\x03\x12\x0f\n\x0b\x44TYPE_INT32\x10\x04\x12\x0f\n\x0b\x44TYPE_INT64\x10\x05\x12\x0f\n\x0b\x44TYPE_UINT8\x10\x06\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "ndarray_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _DTYPE._serialized_start = 227
    _DTYPE._serialized_end = 358
    _ARRAY._serialized_start = 40
    _ARRAY._serialized_end = 224
# @@protoc_insertion_point(module_scope)



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import abc
import copy
import logging
from typing import Awaitable, Callable, Dict, Generic, TypeVar

import cogment
import numpy as np
from cogment.actor import ActorSession
from cogment.environment import EnvironmentSession
from cogment.session import RecvEvent

from cogment_lab.session_helpers import ActorSessionHelper, EnvironmentSessionHelper
from cogment_lab.specs import AgentSpecs


Action = TypeVar("Action")
Actions = Dict[str, Action]

Observation = TypeVar("Observation")
Observations = Dict[str, Observation]

Rewards = Dict[str, float]

Dones = Dict[str, bool]


class State:
    """Class to hold state information."""

    def __setattr__(self, name, value):
        self.__dict__[name] = value

    def __getattr__(self, name):
        """Get attribute from internal dictionary."""
        return self.__dict__.get(name, None)


class BaseEnv(abc.ABC):
    """Base environment class."""

    agent_specs: dict[str, AgentSpecs]

    def __init__(self, *args, **kwargs):
        """Initialize with arguments."""
        self.args = copy.deepcopy(args)
        self.kwargs = copy.deepcopy(kwargs)

    async def impl(self, environment_session: EnvironmentSession):
        """Abstract method to implement environment logic."""
        raise NotImplementedError()

    def get_constructor(self):
        """Get constructor for this environment."""
        cls = self.__class__
        return lambda: cls(*self.args, **self.kwargs)


class CogmentEnv(BaseEnv, abc.ABC, Generic[Observation, Action]):
    """Base Cogment environment class."""

    environment_session: EnvironmentSession
    session_helper: EnvironmentSessionHelper
    agent_specs: dict[str, AgentSpecs]
    actor_name: str

    def __init__(self, *args, **kwargs):
        """Initialize."""
        super().__init__(*args, **kwargs)

    async def initialize(self, state: State, environment_session: EnvironmentSession):
        """Initialize state and session."""
        state.environment_session = environment_session
        return state

    @abc.abstractmethod
    async def reset(self, state: State) -> tuple[State, Observations]:
        """Reset environment state.

        Returns:
            State: Updated state.
            Observations: Initial observations.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def step(self, state: State, action: Actions) -> tuple[State, Observations, Rewards, Dones, Dones, dict]:
        """Take a step in the environment.

        Args:
            state: Current state.
            action: Actions from actors.

        Returns:
            State: Updated state.
            Observations: New observations.
            Rewards: Rewards for each actor.
            Dones: Whether each actor is done.
            Dones: Whether each actor is truncated.
            dict: Additional info.
        """
        raise NotImplementedError()

    async def end(self, state: State):
        """Clean up when done."""
        pass

    async def read_actions(self, state: State, event: RecvEvent):
        """Read actions from event."""
        player_action = state.session_helper.get_action(event, state.actor_name)
        return player_action.value

    async def impl(self, environment_session: EnvironmentSession):
        """Implement environment logic."""
        state = State()
        state = await self.initialize(state, environment_session)
        state, observations = await self.reset(state)

        observations = list(observations.items())

        logging.info("Starting environment session")

        environment_session.start(observations)

        async for event in environment_session.all_events():
            event: RecvEvent
            if event.actions:
                actions = await self.read_actions(state, event)
                (
                    state,
                    observations,
                    rewards,
                    terminateds,
                    truncateds,
                    info,
                ) = await self.step(state, actions)

                dones = {actor_name: terminateds[actor_name] or truncateds[actor_name] for actor_name in terminateds}

                logging.info(f"Adding rewards: {rewards}")
                for actor_name in state.actors:
                    if actor_name not in rewards:
                        rewards[actor_name] = float("nan")
                for actor_name, reward in rewards.items():
                    environment_session.add_reward(value=reward, to=[actor_name], confidence=1.0)

                observations = list(observations.items())

                if all(dones.values()):
                    logging.info("Logging dones=True")
                    environment_session.end(observations)
                # elif event.type != cogment.EventType.ACTIVE:
                #     logging.info("Logging event.type!=ACTIVE")
                #     environment_session.end(observations)
                else:
                    logging.info("Logging a normal observation")
                    environment_session.produce_observations(observations)

        await self.end(state)


class BaseActor(abc.ABC):
    """Base actor class."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        self.args = args
        self.kwargs = kwargs

    async def impl(self, actor_session: ActorSession):
        """Abstract method to implement actor logic."""
        raise NotImplementedError()


class NativeActor(BaseActor):
    """Native actor wrapping a function."""

    def __init__(self, impl: Callable[[ActorSession], Awaitable]):
        """Initialize with implementation function."""
        super().__init__(impl)
        self._impl = impl

    async def impl(self, actor_session: ActorSession):
        """Call implementation function."""
        await self._impl(actor_session)


class CogmentActor(BaseActor, abc.ABC, Generic[Observation, Action]):
    """Base Cogment actor class."""

    actor_session: ActorSession
    current_event: RecvEvent
    session_helper: ActorSessionHelper

    def __init__(self, *args, **kwargs):
        """Initialize."""
        super().__init__(*args, **kwargs)

    async def initialize(self, actor_session: ActorSession):
        """Initialize session and helpers."""
        self.actor_session = actor_session
        self.actor_session.start()

        self.session_helper = ActorSessionHelper(actor_session, None)
        self.action_space = self.session_helper.get_action_space()

    @abc.abstractmethod
    async def act(self, observation: Observation, rendered_frame: np.ndarray | None = None) -> Action:
        """Choose an action based on observation.

        Args:
            observation: Current observation.
            rendered_frame: Optional rendered frame.

        Returns:
            Action to take.
        """
        raise NotImplementedError()

    async def on_reward(self, rewards: list):
        """Handle received rewards."""
        pass

    async def on_message(self, messages: list):
        """Handle received messages."""
        pass

    async def end(self):
        """Clean up when done."""
        pass

    async def impl(self, actor_session: ActorSession):
        """Implement actor logic."""
        await self.initialize(actor_session)
        async for event in actor_session.all_events():
            event: RecvEvent
            self.current_event = event
            if event.type != cogment.EventType.ACTIVE:
                logging.info(f"Skipping event of type {event.type}")
                continue

            if event.observation:
                observation = self.session_helper.get_observation(event)
                if observation is None:
                    continue

                if not observation.active:
                    action = None
                elif not observation.alive:
                    action = None
                else:
                    action = await self.act(observation.value, observation.rendered_frame)  # type: ignore
                cog_action = self.action_space.create_serialize(action)
                actor_session.do_action(cog_action)
            if event.rewards:
                await self.on_reward(event.rewards)
            if event.messages:
                await self.on_message(event.messages)

        await self.end()



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any

import gymnasium as gym
from pettingzoo import AECEnv, ParallelEnv
from pettingzoo.utils.agent_selector import agent_selector


class GymObserverAEC(AECEnv):
    metadata = {"render_modes": ["rgb_array"], "name": "GymWithObserverEnv"}

    def __init__(
        self,
        gym_env_name: str,
        gym_make_kwargs: dict = {},
        render_mode: str | None = None,
    ):
        super().__init__()
        logging.info(
            f"Creating GymObserverAEC with gym_env_name={gym_env_name}, gym_make_kwargs={gym_make_kwargs}, render_mode={render_mode}"
        )
        self.gym_env = gym.make(gym_env_name, render_mode=render_mode, **gym_make_kwargs)
        self.possible_agents = ["gym", "observer"]
        self.observation_spaces = {
            "gym": self.gym_env.observation_space,
            "observer": gym.spaces.Dict(
                {
                    "observation": self.gym_env.observation_space,
                    "action": self.gym_env.action_space,
                }
            ),
        }
        self.action_spaces = {
            "gym": self.gym_env.action_space,
            "observer": self.gym_env.action_space,  # Even though actions are ignored
        }
        self._agent_selector = agent_selector(self.possible_agents)
        # self.reset()

    def reset(self, seed: int | None = None, options: dict | None = None):
        logging.info(f"Resetting GymObserverAEC with seed={seed}, options={options}")
        self.agents = self.possible_agents[:]
        self.agent_selection = self._agent_selector.reset()
        self._cumulative_rewards = {agent: 0 for agent in self.agents}
        self.rewards = {agent: 0 for agent in self.agents}
        self.terminations = {agent: False for agent in self.agents}
        self.truncations = {agent: False for agent in self.agents}
        self._last_observation, info = self.gym_env.reset()
        self.infos = {"gym": info, "observer": info}
        self._gym_action = None

    def step(self, action):
        logging.info(f"Stepping GymObserverAEC with action={action}")
        current_agent = self.agent_selection

        if current_agent == "gym":
            # Execute main agent's action directly
            observation, reward, terminated, truncated, info = self.gym_env.step(action)
            self._last_observation = observation
            self.rewards["gym"] = float(reward)
            self.terminations["gym"] = terminated
            self.truncations["gym"] = truncated
            self.infos["gym"] = info

            self.rewards["observer"] = self.rewards["gym"]
            self.terminations["observer"] = self.terminations["gym"]
            self.truncations["observer"] = self.truncations["gym"]
            self.infos["observer"] = self.infos["gym"]

            self._gym_action = action

        # Observer agent's turn; actions are ignored
        elif current_agent == "observer":
            pass

        self.agent_selection = self._agent_selector.next()

    def observe(self, agent):
        if agent == "gym":
            return self._last_observation
        elif agent == "observer":
            return {"observation": self._last_observation, "action": self._gym_action}

    def render(self):
        return self.gym_env.render()

    def close(self):
        self.gym_env.close()

    def observation_space(self, agent: str):
        return self.observation_spaces[agent]

    def action_space(self, agent: str):
        return self.action_spaces[agent]


class GymObserverParallel(ParallelEnv):
    metadata = {"render_modes": ["rgb_array"], "name": "GymWithObserverEnv"}

    def __init__(
        self,
        gym_env_name: str,
        gym_make_kwargs: dict = {},
        render_mode: str | None = None,
    ):
        super().__init__()
        logging.info(
            f"Creating GymObserverParallel with gym_env_name={gym_env_name}, gym_make_kwargs={gym_make_kwargs}, render_mode={render_mode}"
        )
        self.gym_env = gym.make(gym_env_name, render_mode=render_mode, **gym_make_kwargs)
        self.possible_agents = ["gym", "observer"]
        self.observation_spaces = {
            "gym": self.gym_env.observation_space,
            "observer": self.gym_env.observation_space,
        }
        self.action_spaces = {
            "gym": self.gym_env.action_space,
            "observer": self.gym_env.action_space,  # Even though actions are ignored
        }
        # self.reset()

    def reset(self, seed: int | None = None, options: dict | None = None):
        logging.info(f"Resetting GymObserverParallel with seed={seed}, options={options}")
        self.agents = self.possible_agents[:]

        obs, info = self.gym_env.reset(seed=seed, options=options)

        infos = {"gym": info, "observer": info}
        observations = {"gym": obs, "observer": obs}

        return observations, infos

    def step(self, action: dict[str, Any]):
        logging.info(f"Stepping GymObserverParallel with action={action}")

        obs, reward, terminated, truncated, info = self.gym_env.step(action["gym"])

        observations = {"gym": obs, "observer": obs}
        rewards = {"gym": reward, "observer": reward}
        terminations = {"gym": terminated, "observer": terminated}
        truncations = {"gym": truncated, "observer": truncated}
        infos = {"gym": info, "observer": info}

        return observations, rewards, terminations, truncations, infos

    def render(self):
        return self.gym_env.render()

    def close(self):
        self.gym_env.close()

    def observation_space(self, agent: str):
        return self.observation_spaces[agent]

    def action_space(self, agent: str):
        return self.action_spaces[agent]



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import os
import platform
import re
import stat
from enum import Enum
from tempfile import mkdtemp
from urllib.request import urlopen, urlretrieve


class Arch(Enum):
    AMD64 = "amd64"
    ARM64 = "arm64"


def get_current_arch() -> Arch:
    py_machine = platform.machine()
    if py_machine in ["x86_64", "i686", "AMD64"]:
        return Arch.AMD64

    if py_machine in ["arm64"]:
        return Arch.ARM64

    raise RuntimeError(f"Unsupported architecture [{py_machine}]")


class Os(Enum):
    WINDOWS = "windows"
    LINUX = "linux"
    MACOS = "macos"


def get_current_os() -> Os:
    py_system = platform.system()
    if py_system == "Darwin":
        return Os.MACOS
    if py_system == "Windows":
        return Os.WINDOWS
    if py_system == "Linux":
        return Os.LINUX

    raise RuntimeError(f"Unsupported os [{py_system}]")


def get_latest_release_version() -> str:
    res = urlopen("https://api.github.com/repos/cogment/cogment/releases/latest")

    parsedBody = json.load(res)

    return parsedBody["tag_name"]


def download_cogment(
    output_dir: str | None = None,
    desired_version: str | None = None,
    desired_arch: Arch | None = None,
    desired_os: Os | None = None,
):
    """
    Download a version of cogment

    Parameters:
    - output_dir (string, optional): the output directory, if undefined a temporary directory will be used.
    - desired_version (string, optional): the desired version,
      if undefined the latest released version (excluding prereleases) will be used.
    - desired_arch (Arch, optional): the desired architecture,
      if undefined the current architecture will be detected and used.
    - os (Os, optional): the desired os, if undefined the current os will be detected and used.

    Returns:
        path to the downloaded cogment
    """
    if not output_dir:
        output_dir = mkdtemp()
    else:
        output_dir = os.path.abspath(output_dir)
        os.makedirs(output_dir, exist_ok=True)

    if not desired_version:
        desired_version = get_latest_release_version()

    try:
        desired_version = re.findall(r"[0-9]+.[0-9]+.[0-9]+(?:-[a-zA-Z0-9]+)?", desired_version)[0]
    except RuntimeError as exc:
        raise RuntimeError(f"Desired cogment version [{desired_version}] doesn't follow the expected patterns") from exc

    if desired_arch is None:
        desired_arch = get_current_arch()

    if desired_os is None:
        desired_os = get_current_os()

    cogment_url = (
        "https://github.com/cogment/cogment/releases/download/"
        + f"v{desired_version}/cogment-{desired_os.value}-{desired_arch.value}"
    )

    cogment_filename = os.path.join(output_dir, "cogment")
    if desired_os == Os.WINDOWS:
        cogment_url += ".exe"
        cogment_filename += ".exe"

    try:
        cogment_filename, _ = urlretrieve(cogment_url, cogment_filename)
    except Exception as exc:
        raise RuntimeError(
            f"Unable to retrieve cogment version [{desired_version}] for arch "
            + f"[{desired_arch}] and os [{desired_os}] from [{cogment_url}] to [{cogment_filename}]"
        ) from exc

    # Make sure it is executable
    cogment_stat = os.stat(cogment_filename)
    os.chmod(cogment_filename, cogment_stat.st_mode | stat.S_IEXEC)

    return cogment_filename



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gymnasium as gym
import numpy as np

from cogment_lab.generated.spaces_pb2 import Box  # type: ignore
from cogment_lab.generated.spaces_pb2 import Dict  # type: ignore
from cogment_lab.generated.spaces_pb2 import Discrete  # type: ignore
from cogment_lab.generated.spaces_pb2 import MultiBinary  # type: ignore
from cogment_lab.generated.spaces_pb2 import MultiDiscrete  # type: ignore
from cogment_lab.generated.spaces_pb2 import Space  # type: ignore

from .ndarray_serialization import (
    SerializationFormat,
    deserialize_ndarray,
    serialize_ndarray,
)


def serialize_gym_space(space: gym.Space, serialization_format=SerializationFormat.STRUCTURED) -> Space:
    if isinstance(space, (gym.spaces.Discrete, gym.spaces.Discrete)):
        return Space(discrete=Discrete(n=space.n, start=space.start))
    if isinstance(space, gym.spaces.Box):
        low = space.low
        high = space.high
        return Space(
            box=Box(
                low=serialize_ndarray(low, serialization_format=serialization_format),
                high=serialize_ndarray(high, serialization_format=serialization_format),
            ),
        )

    if isinstance(space, gym.spaces.MultiBinary):
        if isinstance(space.n, np.ndarray):
            size = space.n
        elif isinstance(space.n, int):
            size = np.array([space.n], dtype=np.dtype("int32"))
        else:
            size = np.array(space.n, dtype=np.dtype("int32"))
        return Space(
            multi_binary=MultiBinary(n=serialize_ndarray(size, serialization_format=serialization_format)),
        )

    if isinstance(space, gym.spaces.MultiDiscrete):
        nvec = space.nvec
        return Space(
            multi_discrete=MultiDiscrete(nvec=serialize_ndarray(nvec, serialization_format=serialization_format)),
        )

    if isinstance(space, gym.spaces.Dict):
        spaces = []
        for key, gym_sub_space in space.spaces.items():
            spaces.append(Dict.SubSpace(key=key, space=serialize_gym_space(gym_sub_space)))
        return Space(dict=Dict(spaces=spaces))
    raise RuntimeError(f"[{type(space)}] is not a supported space type")


def deserialize_space(pb_space: Space) -> gym.Space:
    space_kind = pb_space.WhichOneof("kind")
    if space_kind == "discrete":
        discrete_space_pb = pb_space.discrete
        return gym.spaces.Discrete(n=discrete_space_pb.n, start=discrete_space_pb.start)
    if space_kind == "box":
        box_space_pb = pb_space.box
        low = deserialize_ndarray(box_space_pb.low)
        high = deserialize_ndarray(box_space_pb.high)
        assert low is not None
        assert high is not None
        return gym.spaces.Box(low=low, high=high, shape=low.shape, dtype=low.dtype)
    if space_kind == "multi_binary":
        multi_binary_space_pb = pb_space.multi_binary
        size = deserialize_ndarray(multi_binary_space_pb.n)
        assert size is not None
        if size.size > 1:
            return gym.spaces.MultiBinary(n=size)
        return gym.spaces.MultiBinary(n=size[0])
    if space_kind == "multi_discrete":
        multi_discrete_space_pb = pb_space.multi_discrete
        nvec = deserialize_ndarray(multi_discrete_space_pb.nvec)
        assert nvec is not None
        return gym.spaces.MultiDiscrete(nvec=nvec)
    if space_kind == "dict":
        dict_space_pb = pb_space.dict
        spaces = []
        for sub_space in dict_space_pb.spaces:
            spaces.append((sub_space.key, deserialize_space(sub_space.space)))

        return gym.spaces.Dict(spaces=spaces)

    raise RuntimeError(f"[{space_kind}] is not a supported space kind")



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for environment configuration.

Defines data structures for specifying environment configurations.
"""

from __future__ import annotations

from typing import Any, TypedDict


class EnvConfig(TypedDict, total=False):
    """
    Configuration for a single environment.

    Attributes:
        env_type: Type of the environment.
        env_name: Name of the environment.
        cogment_env: Cogment environment ID to use, instead of env_id/registration/make_args.
        env_id: Environment ID.
        registration: Environment registration details.
        make_args: Arguments for make() to construct the environment.
        reset_options: Reset options for the environment.
        render: Whether to render the environment.
    """

    env_type: str
    env_name: str

    # Either this...
    cogment_env: str | None

    # ...or all of this
    env_id: str | None
    registration: str | None
    make_args: dict[str, Any]
    reset_options: dict[str, Any]
    render: bool


class EnvRunnerConfig(TypedDict):
    """
    Configuration for an environment runner.

    Attributes:
        envs: List of EnvConfig, one for each environment.
        port: Port number for the runner.
    """

    envs: list[EnvConfig]
    port: int



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
from typing import Any, TypedDict

import numpy as np
from cogment.environment import EnvironmentSession
from cogment.session import RecvEvent
from pettingzoo import AECEnv, ParallelEnv

from cogment_lab.core import CogmentEnv, State
from cogment_lab.generated.data_pb2 import Observation as PbObservation  # type: ignore
from cogment_lab.session_helpers import EnvironmentSessionHelper
from cogment_lab.specs import AgentSpecs
from cogment_lab.utils import import_object


class PZConfig(TypedDict):
    """Configuration for a PettingZoo environment."""

    env_path: str
    make_args: dict
    reset_options: dict


class AECEnvironment(CogmentEnv):
    """Cogment environment wrapper for PettingZoo AEC environments."""

    def __init__(
        self,
        env_path: str,
        make_kwargs: dict | None = None,
        reset_options: dict | None = None,
        render: bool = False,
        reinitialize: bool = False,
        dry: bool = False,
        sub_dry: bool = True,
    ):
        """
        Initialize the AECEnvironment.

        Args:
            env_path: Path to the PettingZoo environment class.
            make_kwargs: Arguments to pass to the environment constructor.
            reset_options: Options to pass to reset().
            render: Whether to render the environment.
            reinitialize: Whether to reinitialize the environment each session.
            dry: Whether to abstain from initializing the environment in this process.
            sub_dry: Whether to abstain from initializing the environment in the initializer of the subprocess
        """
        super().__init__(
            env_path=env_path,
            make_kwargs=make_kwargs,
            reset_options=reset_options,
            render=render,
            reinitialize=reinitialize,
            dry=sub_dry,
            sub_dry=sub_dry,
        )
        self.env_path = env_path
        self.make_args = make_kwargs or {}
        self.reset_options = reset_options or {}
        self.render = render
        self.reinitialize = reinitialize
        self.dry = dry

        logging.info(
            f"Creating AECEnvironment with {env_path=}, {make_kwargs=}, {reset_options=}, {render=}, {reinitialize=}, {dry=}, {sub_dry=}"
        )

        self.env_maker = import_object(self.env_path)

        assert callable(self.env_maker), f"Environment class at {self.env_path} is not callable"

        if render:
            self.make_args["render_mode"] = "rgb_array"

        if not self.dry:
            self.env: AECEnv = self.env_maker(**self.make_args)
            self.agent_specs = self.create_agent_specs(self.env)
        else:
            self.env = AECEnv()
            self.agent_specs = {}

        self.initialized = False

    async def initialize(self, state: State, environment_session: EnvironmentSession):
        """
        Initialize the environment session.

        Args:
            state: The Cogment state.
            environment_session: The Cogment environment session.

        Returns:
            The updated state.
        """
        logging.info("Initializing environment session")
        if not self.initialized and not self.reinitialize:
            self.env = self.env_maker(**self.make_args)
            self.agent_specs = self.create_agent_specs(self.env)

            state.env = self.env
            state.agent_specs = self.agent_specs
        elif self.initialized and not self.reinitialize:
            state.env = self.env
            state.agent_specs = self.agent_specs
        elif self.reinitialize:
            state.env = self.env_maker(**self.make_args)
            state.agent_specs = self.create_agent_specs(state.env)

        self.initialized = True

        state.environment_session = environment_session
        state.session_helper = EnvironmentSessionHelper(environment_session, state.agent_specs)
        state.session_cfg = state.environment_session.config
        state.actors = state.session_helper.actors

        state.observation_spaces = {agent: state.session_helper.get_observation_space(agent) for agent in state.actors}

        return state

    async def reset(self, state: State):
        """
        Reset the environment.

        Args:
            state: The Cogment state.

        Returns:
            A tuple of (state, observations)
        """
        logging.info("Resetting environment")

        state.env.reset(seed=state.session_cfg.seed)

        obs, reward, term, trunc, info = state.env.last()
        agent = state.env.agent_selection

        state.actor_name = agent

        frame = state.env.render() if state.session_cfg.render else None
        observation = state.observation_spaces[agent].create_serialize(
            value=obs, rendered_frame=frame, active=True, alive=True
        )
        if frame is not None:
            logging.info(f"Frame shape at reset: {frame.shape}")
        else:
            logging.info("Frame at reset is None")

        observations = {agent: observation}
        observations = self.fill_observations_(state=state, observations=observations, frame=frame)

        return state, observations

    async def step(self, state: State, action: Any):
        """
        Take an environment step.

        Args:
            state: The Cogment state.
            action: The action to take.

        Returns:
            A tuple of (state, observations, rewards, terminated, truncated, info)
        """
        logging.info("Stepping environment")

        state.env.step(action)
        obs, reward, terminated, truncated, info = state.env.last()
        agent = state.env.agent_selection

        frame = state.env.render() if state.session_cfg.render else None

        if frame is not None:
            logging.info(f"Frame shape at step: {frame.shape}")
        else:
            logging.info("Frame at step is None")

        observation = state.observation_spaces[agent].create_serialize(
            value=obs,
            rendered_frame=frame,
            active=True,
            alive=not (terminated or truncated),
        )

        # observations = [(agent, observation)]
        observations = {agent: observation}
        rewards = {agent: reward}
        terminateds = state.env.terminations
        truncateds = state.env.truncations

        state.actor_name = agent

        observations = self.fill_observations_(state, observations, frame=frame)

        return state, observations, rewards, terminateds, truncateds, info

    async def end(self, state: State):
        """
        End the environment session.

        Args:
            state: The Cogment state.
        """
        logging.info("Ending environment")
        state.env.close()

    @staticmethod
    def fill_observations_(
        state: State, observations: dict[str, PbObservation], frame: np.ndarray | None
    ) -> dict[str, PbObservation]:
        """
        Fill in any missing observations with the default observation. Mutates the observations dict.

        Args:
            state: The Cogment state.
            observations: The observations dict.
            frame: The rendered frame.

        Returns:
            The filled observations dict.
        """
        if "*" in observations:
            return observations
        for actor_name in state.actors:
            if actor_name not in observations:
                observations[actor_name] = state.observation_spaces[actor_name].create_serialize(
                    rendered_frame=frame, active=False
                )

        return observations

    @staticmethod
    def create_agent_specs(env: AECEnv):
        """
        Create the agent specs from a PettingZoo AEC environment.

        Args:
            env: The PettingZoo AEC environment.

        Returns:
            The agent specs dict.
        """
        # Check all observation and action spaces

        is_homogeneous = True
        observation_space = env.observation_space(env.possible_agents[0])
        action_space = env.action_space(env.possible_agents[0])
        for agent in env.possible_agents:
            if env.observation_space(agent) != observation_space:
                is_homogeneous = False
                break
            if env.action_space(agent) != action_space:
                is_homogeneous = False
                break

        if is_homogeneous:
            one_agent_specs = AgentSpecs.create_homogeneous(
                observation_space=observation_space,
                action_space=action_space,
            )
            agent_specs = {agent: one_agent_specs for agent in env.possible_agents}
        else:
            agent_specs = {
                agent: AgentSpecs.create_homogeneous(
                    observation_space=env.observation_space(agent),
                    action_space=env.action_space(agent),
                )
                for agent in env.possible_agents
            }

        return agent_specs


class ParallelEnvironment(CogmentEnv):
    """Cogment environment wrapper for PettingZoo Parallel environments."""

    def __init__(
        self,
        env_path: str,
        make_kwargs: dict | None = None,
        reset_options: dict | None = None,
        render: bool = False,
        reinitialize: bool = False,
        dry: bool = False,
        sub_dry: bool = True,
    ):
        """
        Initialize the ParallelEnvironment.

        Args:
            env_path: Path to the PettingZoo environment class.
            make_kwargs: Arguments to pass to the environment constructor.
            reset_options: Options to pass to reset().
            render: Whether to render the environment.
            reinitialize: Whether to reinitialize the environment each session.
            dry: Whether to abstain from initializing the environment in this process.
            sub_dry: Whether to abstain from initializing the environment in the initializer of the subprocess
        """
        super().__init__(
            env_path=env_path,
            make_kwargs=make_kwargs,
            reset_options=reset_options,
            render=render,
            reinitialize=reinitialize,
            dry=sub_dry,
            sub_dry=sub_dry,
        )
        self.env_path = env_path
        self.make_args = make_kwargs or {}
        self.reset_options = reset_options or {}
        self.render = render
        self.reinitialize = reinitialize
        self.dry = dry

        logging.info(
            f"Creating ParallelEnvironment with {env_path=}, {make_kwargs=}, {reset_options=}, {render=}, {reinitialize=}, {dry=}, {sub_dry=}"
        )

        self.env_maker = import_object(self.env_path)

        assert callable(self.env_maker), f"Environment class at {self.env_path} is not callable"

        if render:
            self.make_args["render_mode"] = "rgb_array"

        if not self.dry:
            self.env: ParallelEnv = self.env_maker(**self.make_args)
            self.agent_specs = self.create_agent_specs(self.env)
        else:
            self.env = ParallelEnv()
            self.agent_specs = {}

        self.initialized = False

    async def initialize(self, state: State, environment_session: EnvironmentSession):
        """
        Initialize the environment session.

        Args:
            state: The Cogment state.
            environment_session: The Cogment environment session.

        Returns:
            The updated state.
        """
        logging.info("Initializing environment session")
        if not self.initialized and not self.reinitialize:
            self.env: ParallelEnv = self.env_maker(**self.make_args)
            self.agent_specs = self.create_agent_specs(self.env)

            state.env = self.env
            state.agent_specs = self.agent_specs
        elif self.initialized and not self.reinitialize:
            state.env = self.env  # type: ignore
            state.agent_specs = self.agent_specs
        elif self.reinitialize:
            state.env = self.env_maker(**self.make_args)  # type: ignore
            state.agent_specs = self.create_agent_specs(state.env)

        self.initialized = True

        state.environment_session = environment_session
        state.session_helper = EnvironmentSessionHelper(environment_session, state.agent_specs)
        state.session_cfg = state.environment_session.config
        state.actors = state.session_helper.actors

        state.observation_spaces = {agent: state.session_helper.get_observation_space(agent) for agent in state.actors}

        return state

    async def reset(self, state: State):
        """
        Reset the environment.

        Args:
            state: The Cogment state.

        Returns:
            A tuple of (state, observations)
        """
        logging.info("Resetting environment")

        obs, info = state.env.reset(seed=state.session_cfg.seed)

        frame = state.env.render() if state.session_cfg.render else None

        if frame is not None:
            logging.info(f"Frame shape at reset: {frame.shape}")
        else:
            logging.info("Frame at reset is None")

        observations = {
            agent: state.observation_spaces[agent].create_serialize(
                value=obs[agent], rendered_frame=frame, active=True, alive=True
            )
            for agent in obs
        }

        # observations = {agent: observation}
        # observations = self.fill_observations_(state=state, observations=observations, frame=frame)

        return state, observations

    async def step(self, state: State, action: dict[str, Any]):
        """
        Take an environment step.

        Args:
            state: The Cogment state.
            action: The action to take.

        Returns:
            A tuple of (state, observations, rewards, terminated, truncated, info)
        """
        logging.info("Stepping environment")

        obs, rewards, terminated, truncated, info = state.env.step(action)

        frame = state.env.render() if state.session_cfg.render else None

        if frame is not None:
            logging.info(f"Frame shape at step: {frame.shape}")
        else:
            logging.info("Frame at step is None")

        observations = {
            agent: state.observation_spaces[agent].create_serialize(
                value=obs[agent],
                rendered_frame=frame,
                active=True,
                alive=not (terminated[agent] or truncated[agent]),
            )
            for agent in obs
        }

        return state, observations, rewards, terminated, truncated, info

    async def end(self, state: State):
        """
        End the environment session.

        Args:
            state: The Cogment state.
        """
        logging.info("Ending environment")
        state.env.close()

    async def read_actions(self, state: State, event: RecvEvent):
        """Read actions from event."""
        player_actions = {agent: state.session_helper.get_action(event, agent).value for agent in state.actors}
        return player_actions

    @staticmethod
    def create_agent_specs(env: ParallelEnv):
        """
        Create the agent specs from a PettingZoo AEC environment.

        Args:
            env: The PettingZoo AEC environment.

        Returns:
            The agent specs dict.
        """
        # Check all observation and action spaces

        is_homogeneous = True
        observation_space = env.observation_space(env.possible_agents[0])
        action_space = env.action_space(env.possible_agents[0])
        for agent in env.possible_agents:
            if env.observation_space(agent) != observation_space:
                is_homogeneous = False
                break
            if env.action_space(agent) != action_space:
                is_homogeneous = False
                break

        if is_homogeneous:
            one_agent_specs = AgentSpecs.create_homogeneous(
                observation_space=observation_space,
                action_space=action_space,
            )
            agent_specs = {agent: one_agent_specs for agent in env.possible_agents}
        else:
            agent_specs = {
                agent: AgentSpecs.create_homogeneous(
                    observation_space=env.observation_space(agent),
                    action_space=env.action_space(agent),
                )
                for agent in env.possible_agents
            }

        return agent_specs



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from google.protobuf.json_format import ParseDict

from cogment_lab.generated import data_pb2


def extend_actor_config(
    actor_config_template: dict,
    run_id: str,
    agent_specs: data_pb2.AgentSpecs,  # type: ignore
    seed: int,
) -> data_pb2.AgentConfig:  # type: ignore
    """Extends an actor configuration template with additional parameters.

    Args:
        actor_config_template: Template for actor configuration, possibly None.
        run_id: Identifier for the run.
        agent_specs: Specifications for the environment.
        seed: Seed for random number generation.

    Returns:
        An instance of AgentConfig with extended configuration.
    """
    config = data_pb2.AgentConfig()  # type: ignore
    if actor_config_template:
        ParseDict(actor_config_template, config)
    config.run_id = run_id
    config.agent_specs.CopyFrom(agent_specs)
    config.seed = seed
    return config


def create_value(val: str | int | float) -> data_pb2.Value:  # type: ignore
    """Creates a Value protobuf message from a Python value.

    Args:
        val: The input string, integer, or float.

    Returns:
        A Value protobuf message containing the input value.
    """
    value_message = data_pb2.Value()  # type: ignore
    if isinstance(val, str):
        value_message.string_value = val
    elif isinstance(val, int):
        value_message.int_value = val
    elif isinstance(val, float):
        value_message.float_value = val
    else:
        raise ValueError("Unsupported type")
    return value_message


def get_env_config(
    run_id: str | None = None,
    render: bool | None = None,
    render_width: int | None = None,
    seed: int | None = None,
    flatten: bool | None = None,
    reset_args_dict: dict[str, str | int | float] | None = None,
) -> data_pb2.EnvironmentConfig:  # type: ignore
    """Generates an EnvironmentConfig protobuf message.

    Args:
        run_id: Identifier for the run.
        render: Whether to render the environment.
        render_width: Render width if rendering.
        seed: Random seed.
        flatten: Whether to flatten the observation.
        reset_args_dict: Dictionary of reset argument values.

    Returns:
        An EnvironmentConfig protobuf message.
    """
    if reset_args_dict is None:
        reset_args_dict = {}
    env_config = data_pb2.EnvironmentConfig()  # type: ignore

    env_config.run_id = run_id
    env_config.render = render
    env_config.render_width = render_width
    env_config.seed = seed
    env_config.flatten = flatten

    for key, val in reset_args_dict.items():
        env_config.reset_args[key].CopyFrom(create_value(val))

    return env_config



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registers an environment with Cogment and runs the Cogment server"""

from __future__ import annotations

import asyncio
import logging
from multiprocessing import Queue

import cogment

from cogment_lab.core import BaseEnv
from cogment_lab.generated import cog_settings
from cogment_lab.utils.runners import setup_logging


async def register_env(env: BaseEnv, env_name: str, signal_queue: Queue, port: int = 9001):
    """Registers an environment with Cogment and runs the Cogment server

    Args:
        env (BaseEnv): The environment to register
        env_name (str): The name to register the environment under
        signal_queue (Queue): A queue to signal when server has started
        port (int, optional): The port for the Cogment server. Defaults to 9001.

    """
    context = cogment.Context(cog_settings=cog_settings, user_id="cogment_lab")

    context.register_environment(impl=env.impl, impl_name=env_name)
    logging.info(f"Registered environment {env_name} with cogment")

    serve = context.serve_all_registered(cogment.ServedEndpoint(port=port))
    signal_queue.put(True)
    await serve


def env_runner(
    env_class: type,
    env_args: tuple,
    env_kwargs: dict,
    env_name: str,
    signal_queue: Queue,
    port: int = 9001,
    log_file: str | None = None,
):
    """Given an environment, runs it

    Args:
        env_class (type): The environment class to instantiate
        env_args (tuple): Positional arguments for the environment
        env_kwargs (dict): Keyword arguments for the environment
        env_name (str): The name to register the environment under
        signal_queue (Queue): A queue to signal when server has started
        port (int, optional): The port for the Cogment server. Defaults to 9001.
        log_file (str | None, optional): File path to write logs to. Defaults to None.
    """
    if log_file:
        setup_logging(log_file)
    env = env_class(*env_args, **env_kwargs)

    asyncio.run(register_env(env, env_name, signal_queue, port))



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
from multiprocessing import Queue

import cogment

from cogment_lab.core import BaseActor
from cogment_lab.generated import cog_settings
from cogment_lab.utils.runners import setup_logging


async def register_actor(actor: BaseActor, actor_name: str, queue: Queue, port: int = 9002):
    context = cogment.Context(cog_settings=cog_settings, user_id="cogment_lab")

    context.register_actor(impl=actor.impl, impl_name=actor_name, actor_classes=["player"])

    serve = context.serve_all_registered(cogment.ServedEndpoint(port=port))

    queue.put(True)

    await serve


def actor_runner(
    actor_class: type,
    actor_args: tuple,
    actor_kwargs: dict,
    actor_name: str,
    signal_queue: Queue,
    port: int = 9002,
    log_file: str | None = None,
):
    """Given an actor, runs it"""
    if log_file:
        setup_logging(log_file)
    actor = actor_class(*actor_args, **actor_kwargs)

    asyncio.run(register_actor(actor, actor_name, signal_queue, port))



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import io

import numpy as np
from PIL import Image


MAX_RENDERED_WIDTH = 2048


def encode_rendered_frame(rendered_frame: np.ndarray, max_size: int = MAX_RENDERED_WIDTH, format: str = "PNG") -> bytes:
    if max_size <= 0:
        max_size = MAX_RENDERED_WIDTH

    image = Image.fromarray(rendered_frame.astype("uint8"), "RGB")

    width, height = image.size
    if max(height, width) > max_size:
        if height > width:
            new_height = max_size
            new_width = int(new_height / height * width)
        else:
            new_width = max_size
            new_height = int(height / width * new_width)
        image = image.resize((new_width, new_height), Image.ANTIALIAS)  # type: ignore

    with io.BytesIO() as output:
        image.save(output, format=format)
        encoded_frame = output.getvalue()

    return encoded_frame


def decode_rendered_frame(encoded_frame: bytes) -> np.ndarray:
    assert len(encoded_frame) > 0, "Encoded frame is empty"

    with io.BytesIO(encoded_frame) as input:
        image = Image.open(input)
        decoded_frame = np.array(image)

    if decoded_frame is None:
        raise ValueError("Failed to decode the rendered frame.")

    return decoded_frame



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import os
import subprocess
import sys


def setup_logging(log_file: str):
    """
    Set up logging to file and stdout/stderr.

    Args:
        log_file: Path to log file
    """
    # Redirect stdout and stderr to log file
    dirname = os.path.dirname(log_file)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    with open(log_file, "a") as f:
        os.dup2(f.fileno(), sys.stdout.fileno())
        os.dup2(f.fileno(), sys.stderr.fileno())

    # Configure logging
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        handlers=[logging.FileHandler(log_file)],  # , logging.StreamHandler()],
    )


def process_cleanup():
    """
    Clean up any leftover processes related to cogment_lab.
    """
    try:
        pid = os.getpid()

        command = (
            f"ps aux | grep 'cogment-lab' | grep 'multiprocessing' | grep -v grep | "
            f"awk '{{if ($3 != {pid}) print $2}}' | xargs -r kill -9"
        )

        subprocess.run(command, shell=True, check=True)
        print("Processes terminated successfully.")
    except subprocess.CalledProcessError as e:
        print(f"An error occurred: {e}")



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gymnasium as gym
import numpy as np
import yaml


def gym_space_constructors():
    """Registers YAML constructors for Gym spaces.

    This allows Gym spaces to be created automatically from YAML files
    by registering constructors for each space type.
    """

    def box_constructor(loader, node):
        """YAML constructor for Box spaces.

        Args:
            loader: The YAML loader.
            node: The YAML node.

        Returns:
            A Box space constructed from the YAML node.
        """
        values = loader.construct_mapping(node)
        return gym.spaces.Box(
            low=np.array(values.get("low", -np.inf)),
            high=np.array(values.get("high", np.inf)),
            shape=values.get("shape", None),
            dtype=values.get("dtype", np.float32),
            seed=values.get("seed", None),
        )

    def discrete_constructor(loader, node):
        """YAML constructor for Discrete spaces.

        Args:
            loader: The YAML loader.
            node: The YAML node.

        Returns:
            A Discrete space constructed from the YAML node.
        """
        values = loader.construct_mapping(node)
        return gym.spaces.Discrete(n=values["n"], seed=values.get("seed", None), start=values.get("start", 0))

    def multibinary_constructor(loader, node):
        """YAML constructor for MultiBinary spaces.

        Args:
            loader: The YAML loader.
            node: The YAML node.

        Returns:
            A MultiBinary space constructed from the YAML node.
        """
        values = loader.construct_mapping(node)
        return gym.spaces.MultiBinary(n=values["n"], seed=values.get("seed", None))

    def multidiscrete_constructor(loader, node):
        """YAML constructor for MultiDiscrete spaces.

        Args:
            loader: The YAML loader.
            node: The YAML node.

        Returns:
            A MultiDiscrete space constructed from the YAML node.
        """
        values = loader.construct_mapping(node)
        return gym.spaces.MultiDiscrete(
            nvec=np.array(values["nvec"]),
            dtype=values.get("dtype", np.int64),
            seed=values.get("seed", None),
            start=values.get("start", None),
        )

    def text_constructor(loader, node):
        """YAML constructor for Text spaces.

        Args:
            loader: The YAML loader.
            node: The YAML node.

        Returns:
            A Text space constructed from the YAML node.
        """
        values = loader.construct_mapping(node)
        return gym.spaces.Text(
            max_length=values["max_length"],
            min_length=values.get("min_length", 1),
            charset=values.get("charset", "alphanumeric"),
            seed=values.get("seed", None),
        )

    def dict_constructor(loader, node):
        """YAML constructor for Dict spaces.

        Args:
            loader: The YAML loader.
            node: The YAML node.

        Returns:
            A Dict space constructed from the YAML node.
        """
        values = loader.construct_mapping(node)
        spaces = values.pop("spaces", None)
        seed = values.pop("seed", None)
        return gym.spaces.Dict(spaces=spaces, seed=seed, **values)

    def tuple_constructor(loader, node):
        """YAML constructor for Tuple spaces.

        Args:
            loader: The YAML loader.
            node: The YAML node.

        Returns:
            A Tuple space constructed from the YAML node.
        """
        values = loader.construct_sequence(node)
        spaces = values.pop("spaces", None)
        seed = values.pop("seed", None)
        return gym.spaces.Tuple(spaces=spaces, seed=seed)

    def sequence_constructor(loader, node):
        """YAML constructor for Sequence spaces.

        Args:
            loader: The YAML loader.
            node: The YAML node.

        Returns:
            A Sequence space constructed from the YAML node.
        """
        values = loader.construct_mapping(node)
        space = values.get("space")
        seed = values.get("seed", None)
        stack = values.get("stack", False)
        return gym.spaces.Sequence(space=space, seed=seed, stack=stack)

    def graph_constructor(loader, node):
        """YAML constructor for Graph spaces.

        Args:
            loader: The YAML loader.
            node: The YAML node.

        Returns:
            A Graph space constructed from the YAML node.
        """
        values = loader.construct_mapping(node)
        node_space = values.pop("node_space")
        edge_space = values.pop("edge_space", None)
        seed = values.pop("seed", None)
        return gym.spaces.Graph(node_space=node_space, edge_space=edge_space, seed=seed)

    # Register constructors
    yaml.add_constructor("!Box", box_constructor)
    yaml.add_constructor("!Discrete", discrete_constructor)
    yaml.add_constructor("!MultiBinary", multibinary_constructor)
    yaml.add_constructor("!MultiDiscrete", multidiscrete_constructor)
    yaml.add_constructor("!Text", text_constructor)

    yaml.add_constructor("!Dict", dict_constructor)
    yaml.add_constructor("!Tuple", tuple_constructor)
    yaml.add_constructor("!Graph", graph_constructor)
    yaml.add_constructor("!Sequence", sequence_constructor)
    yaml.add_constructor("!Tuple", tuple_constructor)



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: data.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import cogment_lab.generated.ndarray_pb2 as ndarray__pb2
import cogment_lab.generated.spaces_pb2 as spaces__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\ndata.proto\x12\x0b\x63ogment_lab\x1a\rndarray.proto\x1a\x0cspaces.proto"\xd7\x01\n\x10\x45nvironmentSpecs\x12\x16\n\x0eimplementation\x18\x01 \x01(\t\x12\x12\n\nturn_based\x18\x02 \x01(\x08\x12\x13\n\x0bnum_players\x18\x03 \x01(\x05\x12\x34\n\x11observation_space\x18\x04 \x01(\x0b\x32\x19.cogment_lab.spaces.Space\x12/\n\x0c\x61\x63tion_space\x18\x05 \x01(\x0b\x32\x19.cogment_lab.spaces.Space\x12\x1b\n\x13web_components_file\x18\x06 \x01(\t"s\n\nAgentSpecs\x12\x34\n\x11observation_space\x18\x01 \x01(\x0b\x32\x19.cogment_lab.spaces.Space\x12/\n\x0c\x61\x63tion_space\x18\x02 \x01(\x0b\x32\x19.cogment_lab.spaces.Space"Y\n\x05Value\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12\x13\n\tint_value\x18\x02 \x01(\x05H\x00\x12\x15\n\x0b\x66loat_value\x18\x03 \x01(\x02H\x00\x42\x0c\n\nvalue_type"\xf1\x01\n\x11\x45nvironmentConfig\x12\x0e\n\x06run_id\x18\x01 \x01(\t\x12\x0e\n\x06render\x18\x02 \x01(\x08\x12\x14\n\x0crender_width\x18\x03 \x01(\x05\x12\x0c\n\x04seed\x18\x04 \x01(\r\x12\x0f\n\x07\x66latten\x18\x05 \x01(\x08\x12\x41\n\nreset_args\x18\x06 \x03(\x0b\x32-.cogment_lab.EnvironmentConfig.ResetArgsEntry\x1a\x44\n\x0eResetArgsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.cogment_lab.Value:\x02\x38\x01"/\n\nHFHubModel\x12\x0f\n\x07repo_id\x18\x01 \x01(\t\x12\x10\n\x08\x66ilename\x18\x02 \x01(\t"\xa4\x01\n\x0b\x41gentConfig\x12\x0e\n\x06run_id\x18\x01 \x01(\t\x12,\n\x0b\x61gent_specs\x18\x02 \x01(\x0b\x32\x17.cogment_lab.AgentSpecs\x12\x0c\n\x04seed\x18\x03 \x01(\r\x12\x10\n\x08model_id\x18\x04 \x01(\t\x12\x17\n\x0fmodel_iteration\x18\x05 \x01(\x05\x12\x1e\n\x16model_update_frequency\x18\x06 \x01(\x05"\r\n\x0bTrialConfig"\x88\x01\n\x0bObservation\x12*\n\x05value\x18\x01 \x01(\x0b\x32\x1b.cogment_lab.nd_array.Array\x12\x0e\n\x06\x61\x63tive\x18\x02 \x01(\x08\x12\r\n\x05\x61live\x18\x03 \x01(\x08\x12\x1b\n\x0erendered_frame\x18\x04 \x01(\x0cH\x00\x88\x01\x01\x42\x11\n\x0f_rendered_frame":\n\x0cPlayerAction\x12*\n\x05value\x18\x01 \x01(\x0b\x32\x1b.cogment_lab.nd_array.Arrayb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "data_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _ENVIRONMENTCONFIG_RESETARGSENTRY._options = None
    _ENVIRONMENTCONFIG_RESETARGSENTRY._serialized_options = b"8\001"
    _ENVIRONMENTSPECS._serialized_start = 57
    _ENVIRONMENTSPECS._serialized_end = 272
    _AGENTSPECS._serialized_start = 274
    _AGENTSPECS._serialized_end = 389
    _VALUE._serialized_start = 391
    _VALUE._serialized_end = 480
    _ENVIRONMENTCONFIG._serialized_start = 483
    _ENVIRONMENTCONFIG._serialized_end = 724
    _ENVIRONMENTCONFIG_RESETARGSENTRY._serialized_start = 656
    _ENVIRONMENTCONFIG_RESETARGSENTRY._serialized_end = 724
    _HFHUBMODEL._serialized_start = 726
    _HFHUBMODEL._serialized_end = 773
    _AGENTCONFIG._serialized_start = 776
    _AGENTCONFIG._serialized_end = 940
    _TRIALCONFIG._serialized_start = 942
    _TRIALCONFIG._serialized_end = 955
    _OBSERVATION._serialized_start = 958
    _OBSERVATION._serialized_end = 1094
    _PLAYERACTION._serialized_start = 1096
    _PLAYERACTION._serialized_end = 1154
# @@protoc_insertion_point(module_scope)



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import gymnasium as gym

from cogment_lab.generated.data_pb2 import PlayerAction  # type: ignore

from .ndarray_serialization import deserialize_ndarray, serialize_ndarray


# pylint: disable=attribute-defined-outside-init
class Action:
    """
    Cogment Verse actor action

    Properties:
        flat_value:
            The action value, as a flat numpy array.
        value:
            The action value, as a numpy array.
    """

    def __init__(self, gym_space: gym.Space, pb_action=None, value=None):
        """
        Action constructor.
        Shouldn't be called directly, prefer the factory function of ActionSpace.
        """
        self._gym_space = gym_space

        if pb_action is not None:
            assert value is None
            self._pb_action = pb_action
            return

        self._value = value

    def _compute_flat_value(self):
        if hasattr(self, "_value"):
            value = self._value
            if value is None:
                return None
            return gym.spaces.flatten(self._gym_space, self._value)

        if not self._pb_action.value != b"":
            # This happens whenever value is None
            return None

        return deserialize_ndarray(self._pb_action.value)

    @property
    def flat_value(self):
        if not hasattr(self, "_flat_value"):
            self._flat_value = self._compute_flat_value()
        return self._flat_value

    def _compute_value(self):
        flat_value = self.flat_value
        if flat_value is None:
            return None
        return gym.spaces.unflatten(self._gym_space, flat_value)

    @property
    def value(self):
        if not hasattr(self, "_value"):
            self._value = self._compute_value()
        return self._value


class ActionSpace:
    """
    Cogment Verse action space

    Properties:
        gym_space:
            Wrapped Gym space for the action values (cf. https://www.gymlibrary.dev/api/spaces/)
        actor_class:
            Class of the actor for which this space will serialize Action probobug messages
        seed:
            Random seed used when generating random actions
    """

    def __init__(self, gym_space: gym.Space, seed: int | None = None):
        self.gym_space = gym_space

        if seed is not None:
            self.gym_space.seed(int(seed))

    def create(self, value=None):
        """
        Create an Action
        """
        return Action(self.gym_space, value=value)

    def sample(self, mask=None):
        """
        Generate a random Action
        """
        return Action(self.gym_space, value=self.gym_space.sample(mask=mask))

    def serialize(
        self,
        action,
    ):
        """
        Serialize an Action to an Action protobuf message
        """
        if action.value is None:
            return PlayerAction()

        serialized_value = serialize_ndarray(action.flat_value)
        return PlayerAction(value=serialized_value)

    def deserialize(self, pb_action):
        """
        Deserialize an Action from an Action protobuf message
        """
        return Action(self.gym_space, pb_action=pb_action)

    def create_serialize(self, value=None):
        """
        Create and serialize an Action to an Action protobuf message
        """
        return self.serialize(self.create(value))



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import gymnasium as gym
import numpy as np
from pettingzoo import AECEnv, ParallelEnv
from pettingzoo.utils.agent_selector import agent_selector


class GymTeacherAEC(AECEnv):
    metadata = {"render_modes": ["rgb_array"]}

    def __init__(
        self,
        gym_env_name: str,
        gym_make_kwargs: dict = {},
        render_mode: str | None = None,
    ):
        super().__init__()
        self.gym_env = gym.make(gym_env_name, render_mode=render_mode, **gym_make_kwargs)
        self.possible_agents = ["gym", "teacher"]
        self.observation_spaces = {
            "gym": self.gym_env.observation_space,
            "teacher": gym.spaces.Dict(
                {
                    "observation": self.gym_env.observation_space,
                    "action": self.gym_env.action_space,
                }
            ),
        }
        teacher_action_space = gym.spaces.Dict({"active": gym.spaces.Discrete(2), "action": self.gym_env.action_space})
        self.action_spaces = {
            "gym": self.gym_env.action_space,
            "teacher": teacher_action_space,
        }
        self._agent_selector = agent_selector(self.possible_agents)
        self.override = False
        # self.reset()

    def reset(self, seed: int | None = None, options: dict | None = None):
        self.agents = self.possible_agents[:]
        self.agent_selection = self._agent_selector.reset()
        self._cumulative_rewards = {agent: 0 for agent in self.agents}
        self.rewards = {agent: 0 for agent in self.agents}
        self.terminations = {agent: False for agent in self.agents}
        self.truncations = {agent: False for agent in self.agents}
        self._last_observation, info = self.gym_env.reset(seed=seed, options=options)
        self.infos = {"gym": info, "teacher": info}
        self._gym_action = None

    def step(self, action: dict):
        current_agent = self.agent_selection
        next_agent = self._agent_selector.next()

        if current_agent == "gym":
            self._cumulative_rewards["gym"] = 0
            self._cumulative_rewards["teacher"] = 0

            self._gym_action = action

        elif current_agent == "teacher":
            if action["active"] == 1:
                self.override = True
                real_action = action["action"]
            else:
                self.override = False
                real_action = self._gym_action
            observation, reward, terminated, truncated, info = self.gym_env.step(real_action)
            self._last_observation = observation
            self.rewards["gym"] = float(reward)
            self.rewards["teacher"] = float(reward)

            self.terminations["gym"] = terminated
            self.terminations["teacher"] = terminated

            self.truncations["gym"] = truncated
            self.truncations["teacher"] = truncated

            self.infos["gym"] = info
            self.infos["teacher"] = info

            self._accumulate_rewards()
        self.agent_selection = next_agent

    def observe(self, agent):
        if agent == "gym":
            return self._last_observation
        elif agent == "teacher":
            return {"observation": self._last_observation, "action": self._gym_action}

    def render(self):
        img = self.gym_env.render()
        if self.override:
            assert isinstance(
                img, np.ndarray
            ), "Gym environment must return a numpy array when rendering. Make sure you passed render_mode='rgb_array' to the environment."
            W, H, _ = img.shape
            N = 5

            # Set the borders to red
            img[:N, :, :] = [255, 0, 0]  # type: ignore
            img[-N:, :, :] = [255, 0, 0]  # type: ignore
            img[:, :N, :] = [255, 0, 0]  # type: ignore
            img[:, -N:, :] = [255, 0, 0]  # type: ignore
        return img

    def close(self):
        self.gym_env.close()

    def observation_space(self, agent: str):
        return self.observation_spaces[agent]

    def action_space(self, agent: str):
        return self.action_spaces[agent]


class GymTeacherParallel(ParallelEnv):
    metadata = {"render_modes": ["rgb_array"]}

    def __init__(
        self,
        gym_env_name: str,
        gym_make_kwargs: dict = {},
        render_mode: str | None = None,
    ):
        super().__init__()
        self.gym_env = gym.make(gym_env_name, render_mode=render_mode, **gym_make_kwargs)
        self.possible_agents = ["gym", "teacher"]
        self.observation_spaces = {
            "gym": self.gym_env.observation_space,
            "teacher": self.gym_env.observation_space,
        }

        teacher_action_space = gym.spaces.Dict({"active": gym.spaces.Discrete(2), "action": self.gym_env.action_space})
        self.action_spaces = {
            "gym": self.gym_env.action_space,
            "teacher": teacher_action_space,
        }
        self._agent_selector = agent_selector(self.possible_agents)
        self.override = False
        # self.reset()

    def reset(self, seed: int | None = None, options: dict | None = None):
        self.agents = self.possible_agents[:]

        obs, info = self.gym_env.reset(seed=seed, options=options)

        infos = {"gym": info, "teacher": info}
        observations = {"gym": obs, "teacher": obs}

        return observations, infos

    def step(self, action):
        if action["teacher"]["active"] == 1:
            self.override = True
            real_action = action["teacher"]["action"]
        else:
            self.override = False
            real_action = action["gym"]

        observation, reward, terminated, truncated, info = self.gym_env.step(real_action)

        observations = {"gym": observation, "teacher": observation}
        rewards = {"gym": reward, "teacher": reward}
        terminations = {"gym": terminated, "teacher": terminated}
        truncations = {"gym": truncated, "teacher": truncated}
        infos = {"gym": info, "teacher": info}

        return observations, rewards, terminations, truncations, infos

    def render(self):
        img = self.gym_env.render()
        if self.override:
            assert isinstance(
                img, np.ndarray
            ), "Gym environment must return a numpy array when rendering. Make sure you passed render_mode='rgb_array' to the environment."
            W, H, _ = img.shape
            N = 5

            # Set the borders to red
            img[:N, :, :] = [255, 0, 0]  # type: ignore
            img[-N:, :, :] = [255, 0, 0]  # type: ignore
            img[:, :N, :] = [255, 0, 0]  # type: ignore
            img[:, -N:, :] = [255, 0, 0]  # type: ignore
        return img

    def close(self):
        self.gym_env.close()

    def observation_space(self, agent: str):
        return self.observation_spaces[agent]

    def action_space(self, agent: str):
        return self.action_spaces[agent]



# Copyright 2024 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import torch
from coltra import Agent
from coltra.buffers import Action, Observation, OnPolicyRecord

from cogment_lab.utils.trial_utils import TrialData


def convert_trial_data_to_coltra(trial_data: TrialData, agent: Agent) -> OnPolicyRecord:
    """Convert TrialData to OnPolicyRecord.

    Args:
        trial_data (TrialData): TrialData instance
        model (Agent): Model instance to evaluate values

    Returns:
        OnPolicyRecord: Converted OnPolicyRecord instance
    """
    obs = trial_data.observations
    action = trial_data.actions
    reward = trial_data.rewards
    done = trial_data.done
    # state = None  # Assuming 'state' is not provided in TrialData

    # last_value = agent.act(Observation(vector=trial_data.last_observation), get_value=True)[2]["value"]
    # value = agent.act(Observation(vector=trial_data.observations), get_value=True)[2]["value"]

    last_value, _ = agent.value(Observation(vector=trial_data.last_observation), ())  # type: ignore
    value, _ = agent.value(Observation(vector=trial_data.observations), ())  # type: ignore

    last_value = last_value.detach().squeeze(-1).cpu().numpy()
    value = value.detach().squeeze(-1).cpu().numpy()

    # Check if required fields are not None
    if obs is None or action is None or reward is None or done is None:
        raise ValueError("Missing required fields in TrialData for conversion")

    # Create an OnPolicyRecord instance with the mapped fields
    on_policy_record = OnPolicyRecord(
        obs=Observation(vector=obs).tensor(),  # type: ignore
        action=Action(discrete=action).tensor(),  # type: ignore
        reward=torch.tensor(reward.astype(np.float32)),
        value=torch.tensor(value.astype(np.float32)),
        done=torch.tensor(done.astype(np.float32)),
        last_value=torch.tensor(last_value.astype(np.float32)),
    )

    return on_policy_record
